from typing import Self, Tuple
from scipy.linalg import svd, eigh
import numpy as np

from luma.interface.typing import Matrix, Vector
from luma.interface.util import KernelUtil
from luma.core.super import Transformer, Unsupervised, Supervised
from luma.interface.exception import (
    NotFittedError,
    NotConvergedError,
    UnsupportedParameterError,
)


__all__ = (
    "PCA",
    "LDA",
    "KDA",
    "CCA",
    "KernelPCA",
    "TruncatedSVD",
    "FactorAnalysis",
)


class PCA(Transformer, Unsupervised):
    """
    PCA, or Principal Component Analysis, is a dimensionality
    reduction technique. It's primarily used to simplify complex
    data while retaining the most important information.

    Parameters
    ----------
    `n_components` : int, required optional, default=None
        Number of principal components

    """

    def __init__(self, n_components: int | None = None) -> None:
        self.n_components = n_components
        self._fitted = False

        self.set_param_ranges({"n_components": ("0<,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix) -> Self:
        self.mean = np.mean(X, axis=0)
        X_centered = X - self.mean
        covariance_matrix = np.cov(X_centered, rowvar=False)
        eigenvalues, eigenvectors = np.linalg.eigh(covariance_matrix)

        sorted_indices = np.argsort(eigenvalues)[::-1]
        eigenvalues = eigenvalues[sorted_indices]
        eigenvectors = eigenvectors[:, sorted_indices]

        if self.n_components is not None:
            self.eigenvalues = eigenvalues[: self.n_components]
            self.eigenvectors = eigenvectors[:, : self.n_components]
        else:
            self.eigenvalues = eigenvalues
            self.eigenvectors = eigenvectors

        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        X_centered = X - self.mean
        return np.dot(X_centered, self.eigenvectors)

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)


class KernelPCA(Transformer, Unsupervised):
    """
    Kernel PCA is a dimensionality reduction technique that extends
    Principal Component Analysis (PCA) to capture complex, nonlinear
    relationships in data by using a kernel function to transform
    the data into a higher-dimensional space where nonlinear patterns
    can be better captured.

    Parameters
    ----------
    `n_components` : int, required optional, default=None
        Number of principal components
    `deg` : int, default=3
        Polynomial degree of polynomial kernel
    `gamma` : float, default=1.0
        Shape parameter of RBF, sigmoid, and laplacian kernels
    `coef` : float, default=1.0
        Additional coefficient of polynomial and sigmoid kernels
    `kernel` : FuncType, default="rbf"
        Type of kernel functions

    """

    def __init__(
        self,
        n_components: int | None = None,
        deg: int = 3,
        gamma: float = 1.0,
        coef: float = 1.0,
        kernel: KernelUtil.FuncType = "rbf",
    ) -> None:
        self.n_components = n_components
        self.deg = deg
        self.gamma = gamma
        self.coef = coef
        self.kernel = kernel
        self.X = None
        self._fitted = False

        self.set_param_ranges(
            {
                "n_components": ("0<,+inf", int),
                "deg": ("0,+inf", int),
                "gamma": ("0<,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> Self:
        self.X = X
        self._select_kernel_function()
        N = X.shape[0]
        self.K = self._compute_kernel_matrix(X, X)

        one_n = np.ones((N, N)) / N
        self.K_centered = self.K - one_n.dot(self.K) - self.K.dot(one_n)
        self.K_centered += one_n.dot(self.K).dot(one_n)

        self.eigvals, self.eigvecs = eigh(self.K_centered)
        self.eigvals, self.eigvecs = self.eigvals[::-1], self.eigvecs[:, ::-1]

        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)

        K_new = self._compute_kernel_matrix(X, self.X)
        N = self.K.shape[0]
        one_N = np.ones((N, N)) / N
        one_new = np.ones((K_new.shape[0], N)) / N

        K_new_centered = K_new - one_new.dot(self.K)
        K_new_centered += one_new.dot(one_N.dot(self.K))
        K_new_centered -= K_new.dot(one_N.mean(axis=0))

        proj = K_new_centered.dot(self.eigvecs[:, : self.n_components])
        return proj

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)

    def _compute_kernel_matrix(self, X: Matrix, Y: Matrix) -> Matrix:
        N, M = X.shape[0], Y.shape[0]
        K = np.zeros((N, M))
        for i in range(N):
            for j in range(M):
                K[i, j] = self.kernel_func(X[i], Y[j])
        return K

    def _select_kernel_function(self):
        if self.kernel == "linear":
            self.kernel_func = self._linear
        elif self.kernel == "poly":
            self.kernel_func = self._poly
        elif self.kernel == "rbf":
            self.kernel_func = self._rbf
        elif self.kernel == "sigmoid":
            self.kernel_func = self._sigmoid
        elif self.kernel == "laplacian":
            self.kernel_func = self._laplacian
        else:
            raise UnsupportedParameterError(self.kernel)

    def _linear(self, x: Vector, y: Vector) -> float:
        return np.dot(x, y)

    def _poly(self, x: Vector, y: Vector) -> float:
        return (np.dot(x, y) + self.coef) ** self.deg

    def _rbf(self, x: Vector, y: Vector) -> float:
        return np.exp(-self.gamma * np.linalg.norm(x - y) ** 2)

    def _sigmoid(self, x: Vector, y: Vector) -> float:
        return np.tanh(self.gamma * np.dot(x, y) + self.coef)

    def _laplacian(self, x: Vector, y: Vector) -> float:
        return np.exp(-self.gamma * np.linalg.norm(x - y))


class LDA(Transformer, Supervised):
    """
    Linear Discriminant Analysis (LDA) is a dimensionality reduction
    and classification technique. It's primarily employed for dimensionality
    reduction and feature extraction while also considering class
    information for classification tasks.

    Parameters
    ----------
    `n_components` : int, required optional, default=None
        Number of linear discriminants

    Notes
    -----
    * To use LDA for classification, refer to
        `luma.classifier.discriminant.LDAClassifier`

    """

    def __init__(self, n_components: int | None = None) -> None:
        self.n_components = n_components
        self._fitted = False

        self.set_param_ranges({"n_components": ("0<,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Matrix) -> Self:
        self.classes = np.unique(y)
        self.class_means = [np.mean(X[y == c], axis=0) for c in self.classes]

        within_class_scatter = np.zeros((X.shape[1], X.shape[1]))
        for c in self.classes:
            X_c = X[y == c]
            mean_diff = X_c - self.class_means[c]
            within_class_scatter += np.dot(mean_diff.T, mean_diff)

        between_class_scatter = np.zeros((X.shape[1], X.shape[1]))
        for c in self.classes:
            n = X[y == c].shape[0]
            mean_diff = self.class_means[c] - np.mean(X)
            between_class_scatter += n * np.outer(mean_diff, mean_diff)

        eigenvalues, eigenvectors = np.linalg.eigh(
            np.linalg.inv(within_class_scatter).dot(between_class_scatter)
        )

        sorted_indices = np.argsort(eigenvalues)[::-1]
        eigenvalues = eigenvalues[sorted_indices]
        eigenvectors = eigenvectors[:, sorted_indices]

        if self.n_components is not None:
            self.eigenvalues = eigenvalues[: self.n_components]
            self.eigenvectors = eigenvectors[:, : self.n_components]
        else:
            self.eigenvalues = eigenvalues
            self.eigenvectors = eigenvectors

        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return np.dot(X, self.eigenvectors)

    def fit_transform(self, X: Matrix, y: Matrix) -> Matrix:
        self.fit(X, y)
        return self.transform(X)


class TruncatedSVD(Transformer, Unsupervised):
    """
    Truncated Singular Value Decomposition (TruncatedSVD) is a linear dimensionality
    reduction method that simplifies large datasets by preserving the most relevant
    information. It achieves this by decomposing a matrix into three components
    using singular value decomposition (SVD) and retaining only a specified number
    of important components.

    Parameters
    ----------
    `n_components` : int, required optional, default=None
        Dimensionality of low-space

    """

    def __init__(self, n_components: int | None = None) -> None:
        self.n_components = n_components
        self._fitted = False

        self.set_param_ranges({"n_components": ("0<,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix) -> Self:
        mean = np.mean(X, axis=0)
        X_centered = X - mean
        U, S, VT = np.linalg.svd(X_centered, full_matrices=False)

        self.U = U[:, : self.n_components]
        self.S = np.diag(S[: self.n_components])
        self.VT = VT[: self.n_components, :]
        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        mean = np.mean(X, axis=0)
        X_centered = X - mean
        return np.dot(X_centered, self.VT.T)

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)


class FactorAnalysis(Transformer, Unsupervised):
    """
    Factor Analysis is a statistical method used to uncover latent factors
    influencing observed variables. It assumes that observed variables share
    common variance due to these unobservable factors, expressed through
    factor loadings. The technique helps simplify data interpretation by
    revealing the underlying structure of the dataset.

    Parameters
    ----------
    `n_components` : int, required optional, default=None
        Dimensionality of low-space
    `max_iter` : int, default=1000
        Number of iterations
    `tol` : float, default=1e-5
        Threshold for convergence
    `noise_variance` : list or Matirx, optional, default=None
        Initial variances for noise of each features

    """

    def __init__(
        self,
        n_components: int | None = None,
        max_iter: int = 1000,
        tol: float = 1e-5,
        noise_variance: Matrix | list | None = None,
        verbose: bool = False,
    ) -> None:
        self.n_components = n_components
        self.max_iter = max_iter
        self.tol = tol
        self.noise_variance = noise_variance
        self.verbose = verbose
        self._fitted = False

        self.set_param_ranges(
            {"n_components": ("0<,+inf", int), "max_iter": ("0<,+inf", int)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> Self:
        m, n = X.shape
        self.mean = X.mean(axis=0)
        X -= self.mean

        logL_const = n * np.log(2 * np.pi) * self.n_components
        variance = X.var(axis=0)
        psi = np.ones(n)
        if self.noise_variance:
            psi = Matrix(self.noise_variance)

        logLikelihood = []
        prev_logL = -np.inf
        for i in range(self.max_iter):
            sqrt_psi = np.sqrt(psi) + 1e-12
            _, _S, _VT = svd(X / (sqrt_psi * m**0.5), full_matrices=False)
            S, VT = _S[: self.n_components], _VT[: self.n_components]
            unexp_var = np.linalg.norm(S) ** 2

            S **= 2
            W = np.sqrt(np.maximum(S - 1.0, 0.0))[:, np.newaxis] * VT
            W *= sqrt_psi

            logL = np.sum(np.log(S)) + np.sum(np.log(psi))
            logL += logL_const + unexp_var
            logL *= -0.5 * m
            logLikelihood.append(logL)

            abs_diff = abs(logL - prev_logL)
            if abs_diff < self.tol:
                if self.verbose:
                    print(
                        f"[FA] Ealry Convergence at iteration {i}/{self.max_iter}",
                        end=" ",
                    )
                    print(f"with delta-logL of {abs_diff}")
                break
            prev_logL = logL
            psi = np.maximum(variance - np.sum(W**2, axis=0), 1e-12)

            if self.verbose and i % 100 == 0 and i:
                print(f"[FA] Iteration {i}/{self.max_iter} finished", end=" ")
                print(f"with delta-logL of {abs_diff}")

        else:
            NotConvergedError(self)

        self.W = W
        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        Ih = np.eye(len(self.W))
        W_psi = self.W / self.noise_variance if self.noise_variance else self.W
        cov = np.linalg.inv(Ih + W_psi.dot(self.W.T))
        return np.dot((X - self.mean).dot(W_psi.T), cov)

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)


class KDA(Transformer, Supervised):
    """
    Kernel Discriminant Analysis (KDA) is a machine learning technique for
    dimensionality reduction and classification that extends Linear
    Discriminant Analysis (LDA) to nonlinear feature spaces using kernel
    methods. It projects data into a lower-dimensional space where classes
    are more separable by maximizing the ratio of between-class variance
    to within-class variance. KDA utilizes kernel functions to implicitly
    map input data to a high-dimensional space without explicitly computing
    the coordinates in that space.

    Parameters
    ----------
    `n_components` : int, required optional, default=None
        Dimensionality of low-space
    `deg` : int, default=2
        Polynomial degree of polynomial kernel
    `gamma` : float, default=1.0
        Shape parameter of RBF, sigmoid, laplacian
    `coef` : float, default=0.0
        Additional coefficient of polynomial, sigmoid
    `kernel` : FuncType, default="rbf"
        Type of kernel functions

    Notes
    -----
    - To use KDA for classification, refer to
        `luma.classifier.discriminant.KDAClassifier`

    """

    def __init__(
        self,
        n_components: int | None = None,
        deg: int = 2,
        gamma: float = 1.0,
        coef: float = 0.0,
        kernel: KernelUtil.FuncType = "rbf",
    ) -> None:
        self.n_components = n_components
        self.deg = deg
        self.gamma = gamma
        self.coef = coef
        self.kernel = kernel
        self.X_ = None
        self._fitted = False

        self.kernel_params = {
            "deg": self.deg,
            "gamma": self.gamma,
            "coef": self.coef,
        }

        self.set_param_ranges(
            {
                "n_components": ("0<,+inf", int),
                "deg": ("0,+inf", int),
                "gamma": ("0<,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Self:
        m, _ = X.shape
        self.X_ = X
        self.classes = np.unique(y)
        self.ku_ = KernelUtil(self.kernel, **self.kernel_params)

        K = self.ku_.kernel_func(X)
        M = np.mean(K, axis=0)
        Sw = np.zeros((m, m))
        Sb = np.zeros((m, m))

        for i in self.classes:
            Xi = K[y == i]
            Mi = np.mean(Xi, axis=0)
            Ni = Xi.shape[0]

            Sw += np.dot((Xi - Mi).T, Xi - Mi)
            Sb += Ni * np.outer(Mi - M, Mi - M)

        eigvals, eigvecs = np.linalg.eigh(np.linalg.pinv(Sw).dot(Sb))
        indices = np.argsort(eigvals)[::-1]
        self.eigvecs = eigvecs[:, indices[: self.n_components]]

        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Vector:
        K = self.ku_.kernel_func(X, self.X_)
        return K.dot(self.eigvecs)

    def fit_transform(self, X: Matrix, y: Vector) -> Vector:
        self.fit(X, y)
        return self.transform(X)


class CCA(Transformer, Unsupervised):
    """
    Canonical Correlation Analysis (CCA) is a multivariate statistical method
    that finds linear combinations of two sets of variables with the highest
    correlation. It aims to uncover the underlying correlation structure
    between the two variable sets. CCA is an unsupervised technique, often
    used for exploring the relationships in complex data. The result is pairs
    of canonical variables (or components) that represent the maximal
    correlations between the sets.

    Parameters
    ----------
    `n_components` : int, required optional, default=None
        Dimensionality of low-space

    Notes
    -----
    - `CCA` requires two distinct datasets `X` and `Y`,
        in which `Y` is not a target variable
    - Due to its uniqueness in its parameters,
        `CCA` may not be compatible with several meta estimators
    - `transform` and `fit_transform` returns a 2-tuple of `Matrix`

        ```py
        def transform(self, X: Matrix, Y: Matrix) -> Tuple[Matrix, Matrix]
        ```
    """

    def __init__(self, n_components: int | None = None) -> None:
        self.n_components = n_components
        self.correlations_ = None
        self._fitted = False

        self.set_param_ranges({"n_components": ("0<,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix, Y: Matrix) -> Self:
        _, n = X.shape
        X -= X.mean(axis=0)
        Y -= Y.mean(axis=0)

        Cxx = np.cov(X.T)
        Cyy = np.cov(Y.T)
        Cxy = np.cov(X.T, Y.T)[:n, n:]

        inv_Cxx = np.linalg.pinv(Cxx)
        inv_Cyy = np.linalg.pinv(Cyy)
        eigvals, x_weights = np.linalg.eig(inv_Cxx @ Cxy @ inv_Cyy @ Cxy.T)

        indices = eigvals.argsort()[::-1]
        self.x_weights = x_weights[:, indices][:, : self.n_components]
        self.y_weights = inv_Cyy @ Cxy.T @ self.x_weights
        self.y_weights /= eigvals[indices][: self.n_components]

        self.x_scores = X.dot(self.x_weights)
        self.y_scores = Y.dot(self.y_weights)

        self.correlations_ = [
            np.corrcoef(self.x_scores[:, i], self.y_scores[:, i])[0, 1]
            for i in range(self.n_components)
        ]

        self.x_loadings = Cxy.dot(self.y_weights)
        self.y_loadings = Cxy.T.dot(self.x_weights)

        self._fitted = True
        return self

    def transform(self, X: Matrix, Y: Matrix) -> Tuple[Matrix, Matrix]:
        X -= X.mean(axis=0)
        Y -= Y.mean(axis=0)
        X_trans = X.dot(self.x_weights)
        Y_trans = Y.dot(self.y_weights)

        return X_trans, Y_trans

    def fit_transform(self, X: Matrix, Y: Matrix) -> Tuple[Matrix, Matrix]:
        self.fit(X, Y)
        return self.transform(X, Y)



from typing import Tuple
import numpy as np

from luma.core.super import Optimizer
from luma.interface.typing import Tensor, Matrix
from luma.interface.util import InitUtil
from luma.neural.base import Layer


__all__ = ("_Flatten", "_Dense")


class _Flatten(Layer):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, X: Tensor, is_train: bool = False) -> Matrix:
        _ = is_train
        self.input_ = X
        return X.reshape(X.shape[0], -1)

    def backward(self, d_out: Matrix) -> Tensor:
        dX = d_out.reshape(self.input_.shape)
        return dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, *shape = in_shape
        return (batch_size, np.prod(shape))


class _Dense(Layer):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        initializer: InitUtil.InitStr = None,
        optimizer: Optimizer = None,
        lambda_: float = 0.0,
        random_state: int = None,
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.initializer = initializer
        self.optimizer = optimizer
        self.lambda_ = lambda_
        self.random_state = random_state
        self.rs_ = np.random.RandomState(self.random_state)

        self.init_params(
            w_shape=(self.in_features, self.out_features),
            b_shape=(1, self.out_features),
        )
        self.set_param_ranges(
            {
                "in_features": ("0<,+inf", int),
                "out_features": ("0<,+inf", int),
                "lambda_": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    def forward(self, X: Matrix, is_train: bool = False) -> Matrix:
        _ = is_train
        self.input_ = X
        return np.dot(X, self.weights_) + self.biases_

    def backward(self, d_out: Matrix) -> Matrix:
        X = self.input_

        self.dX = np.dot(d_out, self.weights_.T)
        self.dW = np.dot(X.T, d_out)
        self.dW += 2 * self.lambda_ * self.weights_
        self.dB = np.sum(d_out, axis=0, keepdims=True)

        return self.dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, _ = in_shape
        return (batch_size, self.out_features)



from typing import Literal, Self
from scipy.spatial.distance import cdist
from scipy.stats import mode
import numpy as np
import pandas as pd

from luma.interface.typing import Matrix
from luma.core.super import Transformer
from luma.interface.exception import NotFittedError, UnsupportedParameterError


__all__ = ("SimpleImputer", "KNNImputer", "HotDeckImputer")


class SimpleImputer(Transformer, Transformer.Feature):
    def __init__(self, strategy: Literal["mean", "median", "mode"] = "mean"):
        self.strategy = strategy
        self.statistics = None
        self._fitted = False

    def fit(self, X: Matrix) -> Self:
        if self.strategy == "mean":
            self.statistics = np.nanmean(X, axis=0)
        elif self.strategy == "median":
            self.statistics = np.nanmedian(X, axis=0)
        elif self.strategy == "mode":
            self.statistics = []

            for x_col in X.T:
                if np.issubdtype(x_col.dtype, np.number):
                    column_mode, _ = mode(x_col, nan_policy="omit")
                    self.statistics.append(
                        column_mode if column_mode.size > 0 else np.nan
                    )
                else:
                    unique, counts = np.unique(
                        x_col[~pd.isnull(x_col)], return_counts=True
                    )
                    if unique.size > 0:
                        self.statistics.append(unique[np.argmax(counts)])
                    else:
                        self.statistics.append(np.nan)

        else:
            raise UnsupportedParameterError(self.strategy)

        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            NotFittedError(self)
        X_imp = X.copy()
        for i, stat in enumerate(self.statistics):
            if np.issubdtype(X_imp[:, i].dtype, np.number):
                X_imp[:, i] = X_imp[:, i].astype(float)
                mask = np.isnan(X_imp[:, i])
                X_imp[mask, i] = stat
            else:
                mask = pd.isnull(X_imp[:, i]) | (X_imp[:, i] == "nan")
                X_imp[mask, i] = stat

        X_ret = np.zeros_like(X_imp, dtype=object)
        for i in range(X_imp.shape[1]):
            try:
                X_ret[:, i] = X_imp[:, i].astype(float)
            except:
                X_ret[:, i] = X_imp[:, i]

        return X_ret

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)


class KNNImputer(Transformer, Transformer.Feature):
    def __init__(self, n_neighbors: int = 5):
        self.n_neighbors = n_neighbors
        self.distances = None
        self._fitted = False

        self.set_param_ranges({"n_neighbors": ("0<,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix) -> Self:
        if np.isnan(X).any():
            self.distances = self._compute_distances(X)

        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            NotFittedError(self)
        X_imp = X.copy()

        for i, row in enumerate(X):
            nan_indices = np.where(np.isnan(row))[0]
            for x_col in nan_indices:
                neighbor_indices = np.argsort(self.distances[i])
                neighbor_values = []

                for neighbor_index in neighbor_indices:
                    if not np.isnan(X[neighbor_index, x_col]) and neighbor_index != i:
                        neighbor_values.append(X[neighbor_index, x_col])
                        if len(neighbor_values) == self.n_neighbors:
                            break

                if neighbor_values:
                    X_imp[i, x_col] = np.mean(neighbor_values)
                else:
                    X_imp[i, x_col] = np.nanmean(X[:, x_col])

        return X_imp

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)

    def _compute_distances(self, X: Matrix) -> Matrix:
        nan_replacement = np.nanmax(X) + 1
        data_without_nan = np.where(np.isnan(X), nan_replacement, X)
        distances = cdist(data_without_nan, data_without_nan, metric="euclidean")

        return distances


class HotDeckImputer(Transformer, Transformer.Feature):
    def __init__(self):
        self._X = None
        self._similar_rows = None
        self._fitted = False

    def fit(self, X: Matrix) -> Self:
        self._X = X
        self._similar_rows = [self._find_similar_row(row) for row in self._X]

        self._fitted = True
        return self

    def _find_similar_row(self, row: Matrix) -> Matrix:
        min_diff = np.inf
        similar_row = None

        for row in self._X:
            if not np.any(np.isnan(row)):
                diff = np.sum(np.not_equal(row, row) & ~np.isnan(row))
                if diff < min_diff:
                    min_diff = diff
                    similar_row = row

        return similar_row

    def transform(self, X: Matrix) -> Matrix:
        imputed_data = X.copy()
        for i, row in enumerate(imputed_data):
            if np.any(np.isnan(row)) and self._similar_rows[i] is not None:
                missing_indices = np.where(np.isnan(row))[0]
                imputed_data[i, missing_indices] = self._similar_rows[i][
                    missing_indices
                ]

        return imputed_data

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)



import __local__
from luma.neural.layer import Activation

import matplotlib.pyplot as plt
import numpy as np


X = np.linspace(-5, 5, 100)

func = Activation.Swish()
out = func(X)
d_out = func.backward(X)

fig = plt.figure(figsize=(10, 5))
ax1 = fig.add_subplot(1, 2, 1)
ax2 = fig.add_subplot(1, 2, 2)

ax1.plot(X, out, c="dodgerblue", lw=2)
ax1.set_xlabel("x")
ax1.set_ylabel("y")
ax1.set_title(f"{type(func).__name__} Function")
ax1.grid(alpha=0.4)

ax2.plot(X, d_out, c="orange", lw=2)
ax2.set_xlabel("x")
ax2.set_ylabel("y")
ax2.set_title(f"Backward Pass of {type(func).__name__} Function")
ax2.grid(alpha=0.4)

plt.tight_layout()
plt.show()



from typing import Literal, Self, override

from luma.core.super import Estimator, Evaluator, Optimizer, Supervised
from luma.interface.typing import Matrix, Tensor, Vector
from luma.interface.util import InitUtil

from luma.neural.base import Loss, NeuralModel
from luma.neural.block import ConvBlock2D, DenseBlock
from luma.neural.layer import Activation, Dense, Dropout, Flatten, Sequential


__all__ = ("_SimpleMLP", "_SimpleCNN")


class _SimpleMLP(Estimator, Supervised, NeuralModel):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        hidden_layers: list[int] | int,
        *,
        activation: Activation.FuncType,
        optimizer: Optimizer,
        loss: Loss,
        initializer: InitUtil.InitStr = None,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.001,
        valid_size: float = 0.1,
        dropout_rate: float = 0.5,
        lambda_: float = 0.0,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int = None,
        deep_verbose: bool = False,
    ) -> None:
        self.in_features = in_features
        self.out_features = out_features
        self.hidden_layers = hidden_layers
        self.initializer = initializer
        self.activation = activation
        self.optimizer = optimizer
        self.loss = loss
        self.dropout_rate = dropout_rate
        self.lambda_ = lambda_
        self.shuffle = shuffle
        self.random_state = random_state
        self.fitted_ = False

        super().__init__(
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            early_stopping,
            patience,
            deep_verbose,
        )
        super().__init_model__()
        self.model = Sequential()
        self.optimizer.set_params(learning_rate=self.learning_rate)
        self.model.set_optimizer(optimizer=self.optimizer)

        if isinstance(self.hidden_layers, int):
            self.hidden_layers = [self.hidden_layers]

        self.feature_sizes_ = [
            self.in_features,
            *self.hidden_layers,
            self.out_features,
        ]
        self.feature_shapes_ = self._get_feature_shapes(self.feature_sizes_)

        self.set_param_ranges(
            {
                "in_features": ("0<,+inf", int),
                "out_features": ("0<,+inf", int),
                "batch_size": ("0<,+inf", int),
                "n_epochs": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "valid_size": ("0<,<1", None),
                "dropout_rate": ("0,1", None),
                "lambda_": ("0,+inf", None),
                "patience": (f"0<,+inf", int),
            }
        )
        self.check_param_ranges()
        self._build_model()

    def _build_model(self) -> None:
        for i, (in_, out_) in enumerate(self.feature_shapes_):
            self.model += Dense(
                in_,
                out_,
                initializer=self.initializer,
                lambda_=self.lambda_,
                random_state=self.random_state,
            )
            if i < len(self.feature_shapes_) - 1:
                self.model += self.activation()
                self.model += Dropout(
                    dropout_rate=self.dropout_rate,
                    random_state=self.random_state,
                )

    def fit(self, X: Matrix, y: Matrix) -> Self:
        return super(_SimpleMLP, self).fit_nn(X, y)

    @override
    def predict(self, X: Matrix, argmax: bool = True) -> Matrix | Vector:
        return super(_SimpleMLP, self).predict_nn(X, argmax)

    @override
    def score(
        self, X: Matrix, y: Matrix, metric: Evaluator, argmax: bool = True
    ) -> float:
        return super(_SimpleMLP, self).score_nn(X, y, metric, argmax)


class _SimpleCNN(Estimator, Supervised, NeuralModel):
    def __init__(
        self,
        in_channels_list: list[int] | int,
        in_features_list: list[int] | int,
        out_channels: int,
        out_features: int,
        filter_size: int,
        *,
        activation: Activation.FuncType,
        optimizer: Optimizer,
        loss: Loss,
        initializer: InitUtil.InitStr = None,
        padding: Literal["same", "valid"] = "same",
        stride: int = 1,
        do_batch_norm: bool = True,
        momentum: float = 0.9,
        do_pooling: bool = True,
        pool_filter_size: int = 2,
        pool_stride: int = 2,
        pool_mode: Literal["max", "avg"] = "max",
        do_dropout: bool = True,
        dropout_rate: float = 0.5,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.001,
        valid_size: float = 0.1,
        lambda_: float = 0.0,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int = None,
        deep_verbose: bool = False,
    ) -> None:
        self.in_channels_list = in_channels_list
        self.in_features_list = in_features_list
        self.out_channels = out_channels
        self.out_features = out_features
        self.filter_size = filter_size
        self.activation = activation
        self.optimizer = optimizer
        self.loss = loss
        self.initializer = initializer
        self.padding = padding
        self.stride = stride
        self.do_batch_norm = do_batch_norm
        self.momentum = momentum
        self.do_pooling = do_pooling
        self.pool_filter_size = pool_filter_size
        self.pool_stride = pool_stride
        self.pool_mode = pool_mode
        self.do_dropout = do_dropout
        self.dropout_rate = dropout_rate
        self.lambda_ = lambda_
        self.shuffle = shuffle
        self.random_state = random_state
        self._fitted = False

        super().__init__(
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            early_stopping,
            patience,
            deep_verbose,
        )
        super().__init_model__()
        self.model = Sequential()
        self.optimizer.set_params(learning_rate=self.learning_rate)
        self.model.set_optimizer(optimizer=self.optimizer)

        if isinstance(self.in_channels_list, int):
            self.in_channels_list = [self.in_channels_list]
        if isinstance(self.in_features_list, int):
            self.in_features_list = [self.in_features_list]

        self.feature_sizes_ = [
            [*self.in_channels_list, self.out_channels],
            [*self.in_features_list, self.out_features],
        ]
        self.feature_shapes_ = [
            [*self._get_feature_shapes(self.feature_sizes_[0])],
            [*self._get_feature_shapes(self.feature_sizes_[1])],
        ]

        self.set_param_ranges(
            {
                "out_channels": ("0<,+inf", int),
                "out_features": ("0<,+inf", int),
                "filter_size": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
                "momentum": ("0,1", None),
                "pool_filter_size": ("0<,+inf", int),
                "pool_stride": ("0<,+inf", int),
                "dropout_rate": ("0,1", None),
                "batch_size": ("0<,+inf", int),
                "n_epochs": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "valid_size": ("0<,<1", None),
                "dropout_rate": ("0,1", None),
                "lambda_": ("0,+inf", None),
                "patience": (f"0<,+inf", int),
            }
        )
        self.check_param_ranges()
        self._build_model()

    def _build_model(self) -> None:
        for in_, out_ in self.feature_shapes_[0]:
            self.model += ConvBlock2D(
                in_,
                out_,
                self.filter_size,
                activation=self.activation,
                initializer=self.initializer,
                padding=self.padding,
                stride=self.stride,
                lambda_=self.lambda_,
                do_batch_norm=self.do_batch_norm,
                momentum=self.momentum,
                do_pooling=self.do_pooling,
                pool_filter_size=self.pool_filter_size,
                pool_stride=self.pool_stride,
                pool_mode=self.pool_mode,
                random_state=self.random_state,
            )

        self.model += Flatten()
        for i, (in_, out_) in enumerate(self.feature_shapes_[1]):
            if i < len(self.feature_shapes_[1]) - 1:
                self.model += DenseBlock(
                    in_,
                    out_,
                    activation=self.activation,
                    lambda_=self.lambda_,
                    do_dropout=self.do_dropout,
                    dropout_rate=self.dropout_rate,
                    random_state=self.random_state,
                )
            else:
                self.model += Dense(
                    in_,
                    out_,
                    lambda_=self.lambda_,
                    random_state=self.random_state,
                )

    @Tensor.force_dim(4)
    def fit(self, X: Tensor, y: Matrix) -> Self:
        return super(_SimpleCNN, self).fit_nn(X, y)

    @override
    @Tensor.force_dim(4)
    def predict(self, X: Tensor, argmax: bool = True) -> Matrix | Vector:
        return super(_SimpleCNN, self).predict_nn(X, argmax)

    @override
    @Tensor.force_dim(4)
    def score(
        self, X: Tensor, y: Matrix, metric: Evaluator, argmax: bool = True
    ) -> float:
        return super(_SimpleCNN, self).score_nn(X, y, metric, argmax)



from typing import Any, Dict, List, Literal, Self, Tuple
from scipy.special import softmax
import numpy as np

from luma.core.super import Estimator, Evaluator, Supervised
from luma.interface.typing import Matrix, Vector
from luma.interface.util import Clone
from luma.interface.exception import NotFittedError, UnsupportedParameterError

from luma.classifier.tree import DecisionTreeClassifier
from luma.regressor.tree import DecisionTreeRegressor
from luma.preprocessing.encoder import LabelBinarizer
from luma.metric.classification import Accuracy
from luma.metric.regression import MeanSquaredError


__all__ = (
    "AdaBoostClassifier",
    "AdaBoostRegressor",
    "GradientBoostingClassifier",
    "GradientBoostingRegressor",
)


class AdaBoostClassifier(Estimator, Estimator.Meta, Supervised):
    """
    The AdaBoost classifier is an ensemble method that combines multiple
    weak learners, typically decision trees, to form a strong classifier.
    During training, it iteratively adjusts the weights of misclassified
    instances, making them more likely to be correctly predicted in subsequent
    rounds. Each weak learner's contribution to the final prediction is
    weighted based on its accuracy. AdaBoost is adaptive in the sense that
    subsequent weak learners are tweaked in favor of instances that previous
    learners misclassified. The final model makes preds based on the
    weighted majority vote of its weak learners.

    Parameters
    ----------
    `base_estimator` : Estimator, default=DecisionTreeClassifier()
        Base est for training multiple models
    `n_estimators` : int, default=50
        Number of base estimators to fit
    `learning_rate` : float, default=1.0
        Step size of class weights(`alpha`) update
    `**kwargs` : dict[str, Any]
        Additional parameters for base est (i.e. `max_depth`)

    Examples
    --------
    >>> ada_est = AdaBoostClassifier(learning_rate=1.0, ...)
    >>> ada_est.fit(X, y)

    >>> y_pred = ada_est.predict(X)
    >>> est = ada_est[i] # Get i-th est from `ada_est`

    """

    def __init__(
        self,
        base_estimator: Estimator = DecisionTreeClassifier(),
        n_estimators: int = 50,
        learning_rate: float = 1.0,
        verbose: bool = False,
        **kwargs: Dict[str, Any],
    ) -> None:
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.verbose = verbose
        self.classes_ = None
        self.estimators_: List[Estimator] = []
        self.estimator_weights_: List[float] = []
        self.estimator_errors_: List[float] = []
        self._base_estimator_params = kwargs
        self._fitted = False

        self.set_param_ranges(
            {"n_estimators": ("0<,+inf", int), "learning_rate": ("0<,+inf", None)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Matrix) -> Self:
        self.classes_ = np.unique(y)

        m, _ = X.shape
        n_classes = len(self.classes_)
        sample_weights = np.ones(m)

        lb = LabelBinarizer().fit(self.classes_)
        y_bin = lb.transform(y)

        for i in range(self.n_estimators):
            est = Clone(self.base_estimator).get
            if hasattr(est, "max_depth"):
                est.max_depth = 3

            est.set_params(**self._base_estimator_params)
            est.fit(X, y, sample_weights=sample_weights)
            y_pred = est.predict(X)
            y_pred_bin = lb.transform(y_pred)

            error_vect = np.sum(y_bin != y_pred_bin, axis=1) / (n_classes - 1)
            error = np.sum(sample_weights * error_vect) / np.sum(sample_weights)
            error = max(error, 1e-10)
            if error >= (n_classes - 1) / n_classes:
                continue

            alpha = self.learning_rate * np.log((1 - error) / error)
            sample_weights *= np.exp(alpha * error_vect)
            sample_weights /= np.sum(sample_weights)

            self.estimators_.append(est)
            self.estimator_weights_.append(alpha)
            self.estimator_errors_.append(error)

            if self.verbose:
                print(
                    f"[AdaBoost] Finished fitting {type(est).__name__}",
                    f"{i + 1}/{self.n_estimators} with",
                    f"error: {error}, alpha: {alpha}",
                )

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)
        m, _ = X.shape
        est_preds = Matrix([est.predict(X) for est in self.estimators_])
        weighted_preds = np.zeros((m, len(self.classes_)))

        for i, est_pred in enumerate(est_preds):
            for j in range(m):
                cl = np.where(self.classes_ == est_pred[j])[0][0]
                weighted_preds[j, cl] += self.estimator_weights_[i]

            if self.verbose:
                print(
                    f"[AdaBoost] Finished prediction of",
                    f"{type(self.base_estimator).__name__}",
                    f"{i + 1}/{self.n_estimators}",
                )

        return self.classes_[np.argmax(weighted_preds, axis=1)]

    def score(self, X: Matrix, y: Vector, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]


class AdaBoostRegressor(Estimator, Estimator.Meta, Supervised):
    """
    The AdaBoost regressor is an ensemble method that combines multiple
    weak learners, typically decision trees, to form a strong regressor.
    During training, it focuses on reducing the errors of the ensemble by
    adjusting the sample weights and fitting each weak learner to the
    residual errors of the previous learners. The final output is a
    combination of the weak learners' outputs, weighted by their individual
    contributions to the ensemble's performance, which aims to produce
    a more accurate regression model.

    Parameters
    ----------
    `base_estimator` : Estimator, default=DecisionTreeRegressor()
        Base est for training multiple models
    `n_estimators` : int, default=50
        Number of base estimators to fit
    `learning_rate` : float, default=1.0
        Step size of class weights(`alpha`) update
    `loss` : {"linear", "square", "exp"}, default="linear"
        Type of loss function
    `**kwargs` : dict[str, Any]
        Additional parameters for base est (i.e. `max_depth`)

    Examples
    --------
    >>> ada_reg = AdaBoostRegressor(learning_rate=1.0, ...)
    >>> ada_reg.fit(X_train, y_train)

    >>> y_pred = ada_reg.predict(X_test)
    >>> est = ada_reg[i] # Get the i-th est from the `ada_reg`

    """

    def __init__(
        self,
        base_estimator: Estimator = DecisionTreeRegressor(),
        n_estimators: int = 50,
        learning_rate: float = 1.0,
        loss: Literal["linear", "square", "exp"] = "linear",
        verbose: bool = False,
        **kwargs: Dict[str, Any],
    ) -> None:
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.loss = loss
        self.verbose = verbose
        self.estimators_: List[Estimator] = []
        self.estimator_weights_: List[float] = []
        self.estimator_errors_: List[float] = []
        self._base_estimator_params = kwargs
        self._fitted = False

        self.set_param_ranges(
            {"n_estimators": ("0<,+inf", int), "learning_rate": ("0<,+inf", None)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Self:
        m, _ = X.shape
        sample_weights = np.ones(m)

        for i in range(self.n_estimators):
            est = Clone(self.base_estimator).get
            if hasattr(est, "max_depth"):
                est.max_depth = 3

            est.set_params(**self._base_estimator_params)
            est.fit(X, y, sample_weights=sample_weights)
            y_pred = est.predict(X)

            loss, error = self._get_loss(y, y_pred, sample_weights)
            if error == 0 or error >= 1.0:
                continue

            eps = 1e-10
            error = np.maximum(error, eps)
            alpha = self.learning_rate * np.log((1 - error) / error) / 2

            sample_weights *= np.power((1 - error) / error, 1 - loss)
            sample_weights /= np.sum(sample_weights)

            self.estimators_.append(est)
            self.estimator_weights_.append(alpha)
            self.estimator_errors_.append(error)

            if self.verbose:
                print(
                    f"[AdaBoost] Finished fitting {type(est).__name__}",
                    f"{i + 1}/{self.n_estimators} with",
                    f"error: {error}, alpha: {alpha}",
                )

        self._fitted = True
        return self

    def _get_loss(
        self, y_true: Vector, y_pred: Vector, sample_weights: Vector
    ) -> Tuple[Vector, float]:
        if self.loss == "linear":
            loss = np.abs(y_true - y_pred)
        elif self.loss == "square":
            loss = (y_true - y_pred) ** 2
        elif self.loss == "exp":
            loss = 1 - np.exp(-np.abs(y_true - y_pred))
        else:
            raise UnsupportedParameterError(self.loss)

        weighted_error = np.sum(sample_weights * loss) / np.sum(sample_weights)
        return loss, weighted_error

    def predict(self, X: Matrix) -> Vector:
        preds = np.zeros(X.shape[0])
        for i, (est, weight) in enumerate(
            zip(self.estimators_, self.estimator_weights_)
        ):
            preds += weight * est.predict(X)
            if self.verbose:
                print(
                    f"[AdaBoost] Finished prediction of",
                    f"{type(self.base_estimator).__name__}",
                    f"{i + 1}/{self.n_estimators}",
                )

        return preds

    def score(
        self, X: Matrix, y: Vector, metric: Evaluator = MeanSquaredError
    ) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]


class GradientBoostingClassifier(Estimator, Estimator.Meta, Supervised):
    """
    Gradient Boosting Classifier (GBC) is an ensemble learning method that
    builds a series of decision trees in a sequential manner, where each tree
    tries to correct the errors of its predecessor. It combines weak learners
    (typically decision trees) to create a strong predictive model. GBC
    focuses on minimizing a loss function, improving model preds
    iteratively by fitting new models to the residual errors made by previous
    ones. It introduces randomness and prevents overfitting through techniques
    like subsampling and feature selection. GBC is versatile, effective for a
    wide range of classification and regression tasks, and performs well on
    both linear and non-linear problems.

    Parameters
    ----------
    `base_estimator` : Estimator, default=DecisionTreeRegressor()
        Base est for training multiple models
    `n_estimators` : int, default=50
        Number of boosting stages to be run
    `learning_rate` : float, default=0.01
        Shrinking factor of the contribution of each est
    `subsample` : float, default=1.0, range=[0,1]
        Fraction of samples to be used for fitting each est
    `**kwargs` : dict[str, Any]
        Additional parameters for base est (i.e. `max_depth`)

    Notes
    -----
    - Upon using `__getitem__`, it returns a list of base estimators
        of the passed class index:

        ```py
        gb_model = GradientBoostingClassifier(...)
        ...
        gb_model[1] # List of estimators of the class '1'
        gb_model[2][0] # First estimator of the class '2'
        ```
    """

    def __init__(
        self,
        base_estimator: Estimator = DecisionTreeRegressor(),
        n_estimators: int = 50,
        learning_rate: float = 0.01,
        subsample: float = 1.0,
        verbose: bool = False,
        **kwargs: Dict[str, Any],
    ) -> None:
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.subsample = subsample
        self.verbose = verbose
        self.estimators_: List[list] = None
        self.classes_ = None
        self.n_classes_ = None
        self._base_estimator_params = kwargs
        self._fitted = False

        self.set_param_ranges(
            {
                "n_estimators": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "subsample": ("0,1", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Self:
        self._initialize(y)
        m, _ = X.shape
        Fm = np.zeros((m, self.n_classes_)) + self.initial_preds_

        for cl in range(self.n_classes_):
            y_c = (y == self.classes_[cl]).astype(float)
            for i in range(self.n_estimators):
                sample_indices = np.random.choice(
                    np.arange(m), size=int(self.subsample * m), replace=False
                )
                X_sample, y_sample = X[sample_indices], y_c[sample_indices]

                p = softmax(Fm[sample_indices], axis=1)
                residuals = y_sample - p[:, cl]

                est: Estimator = self.estimators_[cl][i]
                if hasattr(est, "max_depth"):
                    est.max_depth = 3

                est.set_params(**self._base_estimator_params)
                est.fit(X_sample, residuals)
                update = est.predict(X)
                Fm[:, cl] += self.learning_rate * update

                if self.verbose:
                    print(
                        f"[GBC] Finished fitting {type(est).__name__}",
                        f"{i + 1}/{self.n_estimators} for class {cl}",
                    )

        self._fitted = True
        return self

    def _initialize(self, y: Vector) -> None:
        self.classes_ = np.unique(y)
        self.n_classes_ = len(self.classes_)
        self.initial_preds_ = np.zeros(self.n_classes_)
        self.estimators_ = [
            [Clone(self.base_estimator).get for _ in range(self.n_estimators)]
            for _ in range(self.n_classes_)
        ]

    def predict_proba(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)
        m, _ = X.shape
        Fm = np.zeros((m, self.n_classes_)) + self.initial_preds_
        for cl in range(self.n_classes_):
            for est in self.estimators_[cl]:
                Fm[:, cl] += self.learning_rate * est.predict(X)

        probs = softmax(Fm, axis=1)
        return probs

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)
        proba = self.predict_proba(X)

        return self.classes_[np.argmax(proba, axis=1)]

    def score(self, X: Matrix, y: Vector, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def __getitem__(self, index: int) -> List[Estimator]:
        return self.estimators_[index]


class GradientBoostingRegressor(Estimator, Estimator.Meta, Supervised):
    """
    Gradient Boosting Regressor (GBR) is an ensemble learning method for
    regression tasks. It builds a series of decision trees in a sequential
    manner, where each tree tries to correct the errors of its predecessor
    by fitting to the residual errors. It combines weak learners to create
    a strong predictive model and minimizes a loss function, often mean
    squared error, iteratively.

    Parameters
    ----------
    `base_estimator` : Estimator, default=DecisionTreeRegressor()
        Base est for training multiple models
    `n_estimators` : int, default=50
        Number of boosting stages to be run
    `learning_rate` : float, default=0.01
        Rate of contribution of each est
    `subsample` : float, default=1.0, range=[0,1]
        Fraction of samples to be used for fitting each est
    `loss` : {"mse", "mae", "huber"}, default="mse"
        Type of loss function
    `delta` : float, default=1.0
        Balancing factor between `mse` and `mae`
    `**kwargs` : dict[str, Any]
        Additional parameters for the base estimator (i.e. `max_depth`)

    """

    def __init__(
        self,
        base_estimator: Estimator = DecisionTreeRegressor(),
        n_estimators: int = 50,
        learning_rate: float = 0.01,
        subsample: float = 1.0,
        loss: Literal["mse", "mae", "huber"] = "mse",
        delta: float = 1.0,
        verbose: bool = False,
        **kwargs: Dict[str, Any],
    ) -> None:
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.subsample = subsample
        self.loss = loss
        self.delta = delta
        self.verbose = verbose
        self.estimators_ = []
        self._base_estimator_params = kwargs
        self._fitted = False

        self.set_param_ranges(
            {
                "n_estimators": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "subsample": ("0,1", None),
                "delta": ("0,1", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Self:
        self.init_pred_ = self._initial_prediction(y)
        self.estimators_ = []
        m, _ = X.shape
        Fm = np.mean(y)

        for i in range(self.n_estimators):
            residuals = self._loss_derivative(y, Fm)
            est = Clone(self.base_estimator).get
            est.set_params(**self._base_estimator_params)

            sample_indices = np.random.choice(
                np.arange(m), size=int(self.subsample * m), replace=False
            )
            X_sample, y_sample = X[sample_indices], residuals[sample_indices]

            est.fit(X_sample, y_sample)
            self.estimators_.append(est)
            Fm += self.learning_rate * est.predict(X)

            if self.verbose:
                print(f"Finished fitting est {i+1}/{self.n_estimators}")

        self._fitted = True
        return self

    def _initial_prediction(self, y: Vector) -> Vector:
        if self.loss in ("mse", "huber"):
            return np.mean(y)
        elif self.loss == "mae":
            return np.median(y)
        else:
            raise UnsupportedParameterError(self.loss)

    def _loss_derivative(self, y_true: Vector, y_pred: Vector) -> Vector:
        if self.loss == "mse":
            return -(y_true - y_pred)
        elif self.loss == "mae":
            return np.sign(y_true - y_pred)
        elif self.loss == "huber":
            diff = y_true - y_pred
            is_small_error = np.abs(diff) <= self.delta
            return np.where(is_small_error, -diff, -self.delta * np.sign(diff))
        else:
            raise UnsupportedParameterError(self.loss)

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)
        preds = sum(self.learning_rate * est.predict(X) for est in self.estimators_)

        return self.init_pred_ + preds

    def score(
        self, X: Matrix, y: Vector, metric: Evaluator = MeanSquaredError
    ) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]



from typing import Tuple
import numpy as np

from luma.interface.typing import Tensor
from luma.neural.base import Layer


__all__ = (
    "_BatchNorm1D",
    "_BatchNorm2D",
    "_BatchNorm3D",
    "_LocalResponseNorm",
    "_LayerNorm",
)


class _BatchNorm1D(Layer):
    def __init__(
        self,
        in_features: int,
        momentum: float = 0.9,
        epsilon: float = 1e-9,
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.momentum = momentum
        self.epsilon = epsilon

        self.gamma = np.ones((1, in_features, 1))
        self.beta = np.zeros((1, in_features, 1))
        self.weights_ = [self.gamma, self.beta]

        self.running_mean = np.zeros((1, in_features, 1))
        self.running_var = np.ones((1, in_features, 1))

    @Tensor.force_dim(3)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        if is_train:
            batch_mean = np.mean(X, axis=(0, 2), keepdims=True)
            batch_var = np.var(X, axis=(0, 2), keepdims=True)

            self.running_mean = (
                self.momentum * self.running_mean + (1 - self.momentum) * batch_mean
            )
            self.running_var = (
                self.momentum * self.running_var + (1 - self.momentum) * batch_var
            )

            self.X_norm = (X - batch_mean) / np.sqrt(batch_var + self.epsilon)
        else:
            self.X_norm = (X - self.running_mean) / np.sqrt(
                self.running_var + self.epsilon
            )

        out = self.weights_[0] * self.X_norm + self.weights_[1]
        return out

    @Tensor.force_dim(3)
    def backward(self, d_out: Tensor) -> Tensor:
        batch_size, _, width = d_out.shape

        dX_norm = d_out * self.weights_[0]
        dgamma = np.sum(d_out * self.X_norm, axis=(0, 2), keepdims=True)
        dbeta = np.sum(d_out, axis=(0, 2), keepdims=True)
        self.dW = [dgamma, dbeta]

        dX = (
            1.0
            / (batch_size * width)
            * np.reciprocal(np.sqrt(self.running_var + self.epsilon))
            * (
                batch_size * width * dX_norm
                - np.sum(dX_norm, axis=(0, 2), keepdims=True)
                - self.X_norm
                * np.sum(dX_norm * self.X_norm, axis=(0, 2), keepdims=True)
            )
        )
        self.dX = dX
        return dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape


class _BatchNorm2D(Layer):
    def __init__(
        self,
        in_features: int,
        momentum: float = 0.9,
        epsilon: float = 1e-9,
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.momentum = momentum
        self.epsilon = epsilon

        gamma_ = np.ones((1, in_features, 1, 1))
        beta_ = np.zeros((1, in_features, 1, 1))
        self.weights_ = [gamma_, beta_]

        self.running_mean = np.zeros((1, in_features, 1, 1))
        self.running_var = np.ones((1, in_features, 1, 1))

    @Tensor.force_dim(4)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        if is_train:
            batch_mean = np.mean(X, axis=(0, 2, 3), keepdims=True)
            batch_var = np.var(X, axis=(0, 2, 3), keepdims=True)

            self.running_mean = (
                self.momentum * self.running_mean + (1 - self.momentum) * batch_mean
            )
            self.running_var = (
                self.momentum * self.running_var + (1 - self.momentum) * batch_var
            )

            self.X_norm = (X - batch_mean) / np.sqrt(batch_var + self.epsilon)
        else:
            self.X_norm = (X - self.running_mean) / np.sqrt(
                self.running_var + self.epsilon
            )

        out = self.weights_[0] * self.X_norm + self.weights_[1]
        return out

    @Tensor.force_dim(4)
    def backward(self, d_out: Tensor) -> Tensor:
        batch_size, _, height, width = d_out.shape
        dX_norm = d_out * self.weights_[0]

        dgamma = np.sum(d_out * self.X_norm, axis=(0, 2, 3), keepdims=True)
        dbeta = np.sum(d_out, axis=(0, 2, 3), keepdims=True)
        self.dW = [dgamma, dbeta]

        dX = (
            (1.0 / (batch_size * height * width))
            * np.reciprocal(np.sqrt(self.running_var + self.epsilon))
            * (
                (batch_size * height * width) * dX_norm
                - np.sum(dX_norm, axis=(0, 2, 3), keepdims=True)
                - self.X_norm
                * np.sum(dX_norm * self.X_norm, axis=(0, 2, 3), keepdims=True)
            )
        )
        self.dX = dX
        return self.dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape


class _BatchNorm3D(Layer):
    def __init__(
        self,
        in_features: int,
        momentum: float = 0.9,
        epsilon: float = 1e-5,
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.momentum = momentum
        self.epsilon = epsilon

        self.gamma = np.ones((1, in_features, 1, 1, 1))
        self.beta = np.zeros((1, in_features, 1, 1, 1))
        self.weights_ = [self.gamma, self.beta]

        self.running_mean = np.zeros((1, in_features, 1, 1, 1))
        self.running_var = np.ones((1, in_features, 1, 1, 1))

    @Tensor.force_dim(5)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        if is_train:
            batch_mean = np.mean(X, axis=(0, 2, 3, 4), keepdims=True)
            batch_var = np.var(X, axis=(0, 2, 3, 4), keepdims=True)

            self.running_mean = (
                self.momentum * self.running_mean + (1 - self.momentum) * batch_mean
            )
            self.running_var = (
                self.momentum * self.running_var + (1 - self.momentum) * batch_var
            )

            self.X_norm = (X - batch_mean) / np.sqrt(batch_var + self.epsilon)
        else:
            self.X_norm = (X - self.running_mean) / np.sqrt(
                self.running_var + self.epsilon
            )

        out = self.weights_[0] * self.X_norm + self.weights_[1]
        return out

    @Tensor.force_dim(5)
    def backward(self, d_out: Tensor) -> Tensor:
        batch_size, _, depth, height, width = d_out.shape

        dX_norm = d_out * self.weights_[0]
        dgamma = np.sum(d_out * self.X_norm, axis=(0, 2, 3, 4), keepdims=True)
        dbeta = np.sum(d_out, axis=(0, 2, 3, 4), keepdims=True)
        self.dW = [dgamma, dbeta]

        dX = (
            1.0
            / (batch_size * depth * height * width)
            * np.reciprocal(np.sqrt(self.running_var + self.epsilon))
            * (
                batch_size * depth * height * width * dX_norm
                - np.sum(dX_norm, axis=(0, 2, 3, 4), keepdims=True)
                - self.X_norm
                * np.sum(dX_norm * self.X_norm, axis=(0, 2, 3, 4), keepdims=True)
            )
        )
        self.dX = dX
        return dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape


class _LocalResponseNorm(Layer):
    def __init__(
        self,
        depth: int,
        alpha: float = 1e-4,
        beta: float = 0.75,
        k: float = 2.0,
    ) -> None:
        super().__init__()
        self.depth = depth
        self.alpha = alpha
        self.beta = beta
        self.k = k

    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X

        channels = X.shape[1]
        sq_input = np.square(X)
        extra_dims = (slice(None),) * (X.ndim - 2)

        scale = self.k + self.alpha * np.array(
            [
                np.sum(
                    sq_input[
                        :,
                        max(0, j - self.depth // 2) : min(
                            channels, j + self.depth // 2 + 1
                        ),
                        *extra_dims,
                    ],
                    axis=1,
                )
                for j in range(channels)
            ]
        ).transpose((1, 0) + tuple(range(2, X.ndim)))

        self.scale = np.power(scale, self.beta)
        return X / self.scale

    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        channels = X.shape[1]

        dX = np.zeros_like(X)
        extra_dims = (slice(None),) * (X.ndim - 2)

        dtmp = -self.beta * np.power(self.scale, -1) * d_out / self.scale
        for i in range(channels):
            dsum = (
                np.sum(
                    dtmp[
                        :,
                        max(0, i - self.depth // 2) : min(
                            channels, i + self.depth // 2 + 1
                        ),
                        *extra_dims,
                    ],
                    axis=1,
                    keepdims=True,
                )
                * 2
                * self.alpha
                * X[:, i, *extra_dims]
            )
            dX[:, i, *extra_dims] = (
                d_out[:, i, *extra_dims] / self.scale[:, i, *extra_dims] + dsum
            )
        self.dX = dX
        return self.dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape


class _LayerNorm(Layer):
    def __init__(self, in_shape: tuple[int] | int, epsilon: float = 1e-5) -> None:
        super().__init__()
        self.in_shape = in_shape
        if isinstance(self.in_shape, int):
            self.in_shape = (self.in_shape,)
        self.epsilon = epsilon

        gamma = np.ones(in_shape[1:])
        beta = np.zeros(in_shape[1:])
        self.weights_ = [gamma, beta]

    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        X_axis = tuple(range(1, X.ndim))

        mean = X.mean(axis=X_axis, keepdims=True)
        var = X.var(axis=X_axis, keepdims=True)

        self.X_norm = (X - mean) / np.sqrt(var + self.epsilon)
        out = self.weights_[0] * self.X_norm + self.weights_[1]
        return out

    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        X_axis = tuple(range(1, X.ndim))
        n_features = np.prod(X.shape[1:])

        X_mu = X - X.mean(axis=X_axis, keepdims=True)
        std_inv = 1.0 / np.sqrt(X.var(axis=X_axis, keepdims=True) + self.epsilon)

        dgamma = np.sum(d_out * self.X_norm, axis=0)
        dbeta = np.sum(d_out, axis=0)
        self.dW = [dgamma, dbeta]

        dX_norm = d_out * self.weights_[0]
        dvar = np.sum(-0.5 * dX_norm * X_mu * std_inv**3, axis=X_axis, keepdims=True)
        dmu = np.sum(dX_norm * -std_inv, axis=X_axis, keepdims=True) + dvar * np.mean(
            -2 * X_mu, axis=X_axis, keepdims=True
        )

        dX = dX_norm * std_inv + (dvar * 2 * X_mu + dmu) / n_features
        self.dX = dX
        return self.dX



import numpy as np

from luma.interface.typing import Matrix
from luma.neural.base import Loss


__all__ = ("CrossEntropy", "BinaryCrossEntropy", "MSELoss")


class CrossEntropy(Loss):
    """
    Combines the softmax activation and cross-entropy loss into
    a single class, which is commonly used for multi-class
    classification problems. This loss function is particularly
    useful when the classes are mutually exclusive.

    This class assumes the true and predicted target values
    are one-hot encoded.
    """

    def __init__(self):
        super().__init__()

    def _softmax(self, y_pred: Matrix) -> Matrix:
        exp_shift = np.exp(y_pred - np.max(y_pred, axis=1, keepdims=True))
        softmax_out = exp_shift / np.sum(exp_shift, axis=1, keepdims=True)

        return softmax_out

    def loss(self, y_true: Matrix, y_pred: Matrix) -> float:
        softmax_probs = self._softmax(y_pred)
        clipped_probs = np.clip(softmax_probs, self.epsilon, 1.0)

        loss = -np.sum(y_true * np.log(clipped_probs)) / y_true.shape[0]
        return loss

    def grad(self, y_true: Matrix, y_pred: Matrix) -> Matrix:
        softmax_probs = self._softmax(y_pred)
        grad = (softmax_probs - y_true) / y_true.shape[0]
        return grad


class BinaryCrossEntropy(Loss):
    """
    Computes the cross entropy loss between true labels and
    predicted probabilities, which is widely used for binary
    classification problems.

    This class assumes the true and predicted target values
    are one-hot encoded.
    """

    def __init__(self) -> None:
        super().__init__()

    def _sigmoid(self, z: Matrix) -> Matrix:
        return 1 / (1 + np.exp(-z))

    def loss(self, y_true: Matrix, y_predZ: Matrix) -> float:
        m = y_true.shape[0]
        y_pred = self._sigmoid(y_predZ)
        y_pred = np.clip(y_pred, self.epsilon, 1 - self.epsilon)

        loss = -np.sum(y_true * np.log(y_pred) + (1 - y_true) * np.log(1 - y_pred))
        loss /= m
        return loss

    def grad(self, y_true: Matrix, y_predZ: Matrix) -> Matrix:
        m = y_true.shape[0]
        y_pred = self._sigmoid(y_predZ)
        return (y_pred - y_true) / m


class MSELoss(Loss):
    """
    Calculates the mean squared error between the predicted values
    and the true values, typically used for regression problems.
    Mean squared error is a common measure of the accuracy of a regression
    model and provides a simple way to quantify the difference between
    predicted and actual values.

    Notes
    -----
    - This class is for the loss of neural networks, which differs from
        that of statistical ML models.

    - For statistical models, please refer to
        `luma.metric.regression.MeanSquaredError`.

    """

    def __init__(self) -> None:
        super().__init__()

    def loss(self, y_true: Matrix, y_pred: Matrix) -> float:
        m = y_true.shape[0]
        loss = np.sum((y_pred - y_true) ** 2) / m
        return loss

    def grad(self, y_true: Matrix, y_pred: Matrix) -> Matrix:
        m = y_true.shape[0]
        return 2 * (y_pred - y_true) / m



from typing import Literal, Tuple
import numpy as np

from luma.core.super import Optimizer
from luma.interface.typing import Tensor
from luma.interface.util import InitUtil
from luma.interface.exception import UnsupportedParameterError
from luma.neural.base import Layer


__all__ = ("_Conv1D", "_Conv2D", "_Conv3D")


class _Conv1D(Layer):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        filter_size: Tuple[int] | int,
        stride: int = 1,
        padding: Tuple[int] | int | Literal["valid", "same"] = "same",
        initializer: InitUtil.InitStr = None,
        optimizer: Optimizer | None = None,
        lambda_: float = 0.0,
        random_state: int | None = None,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride
        self.padding = padding
        self.initializer = initializer
        self.optimizer = optimizer
        self.lambda_ = lambda_
        self.random_state = random_state
        self.rs_ = np.random.RandomState(self.random_state)

        if isinstance(filter_size, int):
            self.filter_size = (filter_size,)
        else:
            self.filter_size = filter_size

        self.init_params(
            w_shape=(
                self.out_channels,
                self.in_channels,
                *self.filter_size,
            ),
            b_shape=(1, self.out_channels),
        )
        self.set_param_ranges(
            {
                "in_channels": ("0<,+inf", int),
                "out_channels": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
                "lambda_": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(3)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, width = X.shape

        if self.in_channels != channels:
            raise ValueError(
                f"channels of 'X' does not match with 'in_channels'! "
                + f"({self.in_channels}!={channels})"
            )

        pad_w, padded_w = self._get_padding_dim(width)

        out_width = ((padded_w - self.filter_size[0]) // self.stride) + 1
        out: Tensor = np.zeros((batch_size, self.out_channels, out_width))

        X_padded = np.pad(X, ((0, 0), (0, 0), (pad_w, pad_w)), mode="constant")
        X_fft = np.fft.rfft(X_padded, n=padded_w, axis=2)
        filter_fft = np.fft.rfft(
            self.weights_,
            n=padded_w,
            axis=2,
        )

        for i in range(batch_size):
            for f in range(self.out_channels):
                result_fft = np.sum(X_fft[i] * filter_fft[f], axis=0)
                result = np.fft.irfft(result_fft, n=padded_w)

                out[i, f] = result[pad_w : padded_w - pad_w : self.stride][:out_width]

        out += self.biases_[:, :, np.newaxis]
        return out

    @Tensor.force_dim(3)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        batch_size, channels, width = X.shape
        pad_w, padded_w = self._get_padding_dim(width)

        dX_padded = np.zeros((batch_size, channels, padded_w))
        self.dW = np.zeros_like(self.weights_)
        self.dB = np.zeros_like(self.biases_)

        X_padded = np.pad(X, ((0, 0), (0, 0), (pad_w, pad_w)), mode="constant")
        X_fft = np.fft.rfft(X_padded, n=padded_w, axis=2)
        d_out_fft = np.fft.rfft(d_out, n=padded_w, axis=2)

        for f in range(self.out_channels):
            self.dB[:, f] = np.sum(d_out[:, f, :])

        for f in range(self.out_channels):
            for c in range(channels):
                filter_d_out_fft = np.sum(
                    X_fft[:, c] * d_out_fft[:, f].conj(),
                    axis=0,
                )
                self.dW[f, c] = np.fft.irfftn(
                    filter_d_out_fft,
                    s=(padded_w,),
                )[pad_w : pad_w + self.filter_size[0]]

        self.dW += 2 * self.lambda_ * self.weights_

        for i in range(batch_size):
            for c in range(channels):
                temp = np.zeros(padded_w // 2 + 1, dtype=np.complex128)
                for f in range(self.out_channels):
                    filter_fft = np.fft.rfft(self.weights_[f, c], n=padded_w)
                    temp += filter_fft * d_out_fft[i, f]
                dX_padded[i, c] = np.fft.irfft(temp, n=padded_w)

        self.dX = dX_padded[:, :, pad_w:-pad_w] if pad_w > 0 else dX_padded
        return self.dX

    def _get_padding_dim(self, width: int) -> Tuple[int, ...]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 1:
                raise ValueError("Padding tuple must have exactly one value.")
            pad_w = self.padding[0]

        elif isinstance(self.padding, int):
            pad_w = self.padding
        elif self.padding == "same":
            pad_w = (self.filter_size[0] - 1) // 2
        elif self.padding == "valid":
            pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_w = width + 2 * pad_w
        return pad_w, padded_w

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, _, width = in_shape
        _, padded_w = self._get_padding_dim(width)
        out_width = ((padded_w - self.filter_size[0]) // self.stride) + 1

        return (batch_size, self.out_channels, out_width)


class _Conv2D(Layer):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        filter_size: Tuple[int, int] | int,
        stride: int = 1,
        padding: Tuple[int, int] | int | Literal["valid", "same"] = "same",
        initializer: InitUtil.InitStr = None,
        optimizer: Optimizer | None = None,
        lambda_: float = 0.0,
        random_state: int | None = None,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride
        self.padding = padding
        self.initializer = initializer
        self.optimizer = optimizer
        self.lambda_ = lambda_
        self.random_state = random_state
        self.rs_ = np.random.RandomState(self.random_state)

        if isinstance(filter_size, int):
            self.filter_size = (filter_size, filter_size)
        else:
            self.filter_size = filter_size

        self.init_params(
            w_shape=(
                self.out_channels,
                self.in_channels,
                *self.filter_size,
            ),
            b_shape=(1, self.out_channels),
        )
        self.set_param_ranges(
            {
                "in_channels": ("0<,+inf", int),
                "out_channels": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
                "lambda_": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(4)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, height, width = X.shape

        if self.in_channels != channels:
            raise ValueError(
                f"channels of 'X' does not match with 'in_channels'! "
                + f"({self.in_channels}!={channels})"
            )

        pad_h, pad_w, padded_h, padded_w = self._get_padding_dim(height, width)

        out_height = ((padded_h - self.filter_size[0]) // self.stride) + 1
        out_width = ((padded_w - self.filter_size[1]) // self.stride) + 1
        out: Tensor = np.zeros(
            (
                batch_size,
                self.out_channels,
                out_height,
                out_width,
            )
        )

        X_padded = np.pad(
            X, ((0, 0), (0, 0), (pad_h, pad_h), (pad_w, pad_w)), mode="constant"
        )
        X_fft = np.fft.rfftn(X_padded, s=(padded_h, padded_w), axes=[2, 3])
        filter_fft = np.fft.rfftn(
            self.weights_,
            s=(padded_h, padded_w),
            axes=[2, 3],
        )
        for i in range(batch_size):
            for f in range(self.out_channels):
                result_fft = np.sum(X_fft[i] * filter_fft[f], axis=0)
                result = np.fft.irfftn(result_fft, s=(padded_h, padded_w))

                sampled_result = result[
                    pad_h : padded_h - pad_h : self.stride,
                    pad_w : padded_w - pad_w : self.stride,
                ]
                out[i, f] = sampled_result[:out_height, :out_width]

        out += self.biases_[:, :, np.newaxis, np.newaxis]
        return out

    @Tensor.force_dim(4)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        batch_size, channels, height, width = X.shape
        pad_h, pad_w, padded_h, padded_w = self._get_padding_dim(height, width)

        dX_padded = np.zeros((batch_size, channels, padded_h, padded_w))
        self.dW = np.zeros_like(self.weights_)
        self.dB = np.zeros_like(self.biases_)

        X_padded = np.pad(
            X, ((0, 0), (0, 0), (pad_h, pad_h), (pad_w, pad_w)), mode="constant"
        )
        X_fft = np.fft.rfftn(X_padded, s=(padded_h, padded_w), axes=[2, 3])
        d_out_fft = np.fft.rfftn(d_out, s=(padded_h, padded_w), axes=[2, 3])

        for f in range(self.out_channels):
            self.dB[:, f] = np.sum(d_out[:, f, :, :])

        for f in range(self.out_channels):
            for c in range(channels):
                filter_d_out_fft = np.sum(
                    X_fft[:, c] * d_out_fft[:, f].conj(),
                    axis=0,
                )
                self.dW[f, c] = np.fft.irfftn(
                    filter_d_out_fft,
                    s=(padded_h, padded_w),
                )[
                    pad_h : pad_h + self.filter_size[0],
                    pad_w : pad_w + self.filter_size[1],
                ]

        self.dW += 2 * self.lambda_ * self.weights_

        for i in range(batch_size):
            for c in range(channels):
                temp = np.zeros((padded_h, padded_w // 2 + 1), dtype=np.complex128)
                for f in range(self.out_channels):
                    filter_fft = np.fft.rfftn(
                        self.weights_[f, c], s=(padded_h, padded_w)
                    )
                    temp += filter_fft * d_out_fft[i, f]
                dX_padded[i, c] = np.fft.irfftn(temp, s=(padded_h, padded_w))

        self.dX = (
            dX_padded[:, :, pad_h:-pad_h, pad_w:-pad_w]
            if pad_h > 0 or pad_w > 0
            else dX_padded
        )
        return self.dX

    def _get_padding_dim(self, height: int, width: int) -> Tuple[int, ...]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 2:
                raise ValueError("Padding tuple must have exactly two values.")
            pad_h, pad_w = self.padding
        elif isinstance(self.padding, int):
            pad_h = pad_w = self.padding

        elif self.padding == "same":
            pad_h = (self.filter_size[0] - 1) // 2
            pad_w = (self.filter_size[1] - 1) // 2
        elif self.padding == "valid":
            pad_h = pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_h = height + 2 * pad_h
        padded_w = width + 2 * pad_w

        return pad_h, pad_w, padded_h, padded_w

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, _, height, width = in_shape
        _, _, padded_h, padded_w = self._get_padding_dim(height, width)

        out_height = ((padded_h - self.filter_size[0]) // self.stride) + 1
        out_width = ((padded_w - self.filter_size[1]) // self.stride) + 1

        return (batch_size, self.out_channels, out_height, out_width)


class _Conv3D(Layer):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        filter_size: Tuple[int, int, int] | int,
        stride: int = 1,
        padding: Tuple[int, int, int] | int | Literal["valid", "same"] = "same",
        initializer: InitUtil.InitStr = None,
        optimizer: Optimizer | None = None,
        lambda_: float = 0.0,
        random_state: int | None = None,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride
        self.padding = padding
        self.initializer = initializer
        self.optimizer = optimizer
        self.lambda_ = lambda_
        self.random_state = random_state
        self.rs_ = np.random.RandomState(self.random_state)

        if isinstance(filter_size, int):
            self.filter_size = (filter_size, filter_size, filter_size)
        else:
            self.filter_size = filter_size

        self.init_params(
            w_shape=(
                self.out_channels,
                self.in_channels,
                *self.filter_size,
            ),
            b_shape=(1, self.out_channels),
        )
        self.set_param_ranges(
            {
                "in_channels": ("0<,+inf", int),
                "out_channels": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
                "lambda_": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(5)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, depth, height, width = X.shape

        if self.in_channels != channels:
            raise ValueError(
                f"channels of 'X' does not match with 'in_channels'! "
                + f"({self.in_channels}!={channels})"
            )

        pad_d, pad_h, pad_w, padded_d, padded_h, padded_w = self._get_padding_dim(
            depth, height, width
        )
        out_depth = ((padded_d - self.filter_size[0]) // self.stride) + 1
        out_height = ((padded_h - self.filter_size[1]) // self.stride) + 1
        out_width = ((padded_w - self.filter_size[2]) // self.stride) + 1
        out: Tensor = np.zeros(
            (batch_size, self.out_channels, out_depth, out_height, out_width)
        )

        X_padded = np.pad(
            X,
            ((0, 0), (0, 0), (pad_d, pad_d), (pad_h, pad_h), (pad_w, pad_w)),
            mode="constant",
        )
        X_fft = np.fft.rfftn(
            X_padded,
            s=(padded_d, padded_h, padded_w),
            axes=[2, 3, 4],
        )
        filter_fft = np.fft.rfftn(
            self.weights_,
            s=(padded_d, padded_h, padded_w),
            axes=[2, 3, 4],
        )

        for i in range(batch_size):
            for f in range(self.out_channels):
                result_fft = np.sum(X_fft[i] * filter_fft[f], axis=0)
                result = np.fft.irfftn(result_fft, s=(padded_d, padded_h, padded_w))

                sampled_result = result[
                    pad_d : padded_d - pad_d : self.stride,
                    pad_h : padded_h - pad_h : self.stride,
                    pad_w : padded_w - pad_w : self.stride,
                ]
                out[i, f] = sampled_result[:out_depth, :out_height, :out_width]

        out += self.biases_[:, :, np.newaxis, np.newaxis, np.newaxis]
        return out

    @Tensor.force_dim(5)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        batch_size, channels, depth, height, width = X.shape
        pad_d, pad_h, pad_w, padded_d, padded_h, padded_w = self._get_padding_dim(
            depth, height, width
        )

        dX_padded = np.zeros((batch_size, channels, padded_d, padded_h, padded_w))
        self.dW = np.zeros_like(self.weights_)
        self.dB = np.zeros_like(self.biases_)

        X_padded = np.pad(
            X,
            ((0, 0), (0, 0), (pad_d, pad_d), (pad_h, pad_h), (pad_w, pad_w)),
            mode="constant",
        )
        X_fft = np.fft.rfftn(
            X_padded,
            s=(padded_d, padded_h, padded_w),
            axes=[2, 3, 4],
        )
        d_out_fft = np.fft.rfftn(
            d_out,
            s=(padded_d, padded_h, padded_w),
            axes=[2, 3, 4],
        )

        for f in range(self.out_channels):
            self.dB[:, f] = np.sum(d_out[:, f, :, :, :])

        for f in range(self.out_channels):
            for c in range(channels):
                filter_d_out_fft = np.sum(
                    X_fft[:, c] * d_out_fft[:, f].conj(),
                    axis=0,
                )
                self.dW[f, c] = np.fft.irfftn(
                    filter_d_out_fft,
                    s=(padded_d, padded_h, padded_w),
                )[
                    pad_d : pad_d + self.filter_size[0],
                    pad_h : pad_h + self.filter_size[1],
                    pad_w : pad_w + self.filter_size[2],
                ]

        self.dW += 2 * self.lambda_ * self.weights_

        for i in range(batch_size):
            for c in range(channels):
                temp = np.zeros(
                    (padded_d, padded_h, padded_w // 2 + 1), dtype=np.complex128
                )
                for f in range(self.out_channels):
                    filter_fft = np.fft.rfftn(
                        self.weights_[f, c],
                        s=(padded_d, padded_h, padded_w),
                        axes=[2, 3, 4],
                    )
                    temp += filter_fft * d_out_fft[i, f]
                dX_padded[i, c] = np.fft.irfftn(
                    temp,
                    s=(padded_d, padded_h, padded_w),
                )

        self.dX = (
            dX_padded[:, :, pad_d:-pad_d, pad_h:-pad_h, pad_w:-pad_w]
            if (pad_d > 0 or pad_h > 0 or pad_w > 0)
            else dX_padded
        )
        return self.dX

    def _get_padding_dim(
        self,
        depth: int,
        height: int,
        width: int,
    ) -> Tuple[int, ...]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 3:
                raise ValueError("Padding tuple must have exactly three values.")
            pad_d, pad_h, pad_w = self.padding

        elif isinstance(self.padding, int):
            pad_d = pad_h = pad_w = self.padding
        elif self.padding == "same":
            pad_d = (self.filter_size[0] - 1) // 2
            pad_h = (self.filter_size[1] - 1) // 2
            pad_w = (self.filter_size[2] - 1) // 2
        elif self.padding == "valid":
            pad_d = pad_h = pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_d = depth + 2 * pad_d
        padded_h = height + 2 * pad_h
        padded_w = width + 2 * pad_w

        return pad_d, pad_h, pad_w, padded_d, padded_h, padded_w

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, _, depth, height, width = in_shape
        _, _, _, padded_d, padded_h, padded_w = self._get_padding_dim(
            depth, height, width
        )
        out_depth = ((padded_d - self.filter_size[0]) // self.stride) + 1
        out_height = ((padded_h - self.filter_size[1]) // self.stride) + 1
        out_width = ((padded_w - self.filter_size[2]) // self.stride) + 1

        return (batch_size, self.out_channels, out_depth, out_height, out_width)



from typing import Literal, Self, Tuple
import numpy as np

from luma.interface.typing import Matrix, Vector
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.core.super import Estimator, Evaluator, Supervised
from luma.interface.util import DecisionTreeNode
from luma.metric.classification import Accuracy


__all__ = "DecisionTreeClassifier"


class DecisionTreeClassifier(Estimator, Supervised):
    """
    A Decision Tree Classifier is a machine learning algorithm that
    classifies data by making decisions based on asking a series of
    questions. It splits the dataset into branches at each node,
    based on feature values, aiming to increase homogeneity in each
    branch. The process continues recursively until it reaches a
    leaf node, where the final decision or classification is made.
    This model is intuitive and can easily handle categorical and
    numerical data.

    Parameters
    ----------
    `max_depth` : int, default=10
        Maximum depth of the tree
    `criterion` : {"gini", "entropy"}, default="gini"
        Function used to measure the quality of a split
    `min_samples_split` : int, default=2
        Minimum samples required to split a node
    `min_samples_leaf` : int, default=1
        Minimum samples required to be at a leaf node
    `max_features` : int, optional, default=None
        Number of features to consider
    `min_impurity_decrease` : float, default=0.0
        Minimum decrement of impurity for a split
    `max_leaf_nodes` : int, optional, default=None
        Maximum amount of leaf nodes
    `random_state` : int, optional, default=None
        The randomness seed of the estimator

    """

    def __init__(
        self,
        max_depth: int = 10,
        criterion: Literal["gini", "entropy"] = "gini",
        min_samples_split: int = 2,
        min_samples_leaf: int = 1,
        max_features: int | None = None,
        min_impurity_decrease: float = 0.0,
        max_leaf_nodes: int | None = None,
        random_state: int | None = None,
    ) -> None:
        self.max_depth = max_depth
        self.criterion = criterion
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.min_impurity_decrease = min_impurity_decrease
        self.max_leaf_nodes = max_leaf_nodes
        self.random_state = random_state
        self.root = None
        self.classes_ = None
        self._fitted = False

        self.set_param_ranges(
            {
                "max_depth": ("0<,+inf", int),
                "min_samples_split": ("0,+inf", int),
                "min_samples_leaf": ("0,+inf", int),
                "min_impurity_decrease": ("0,+inf", None),
                "max_leaf_nodes": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()
        np.random.seed(random_state)

    def fit(self, X: Matrix, y: Vector, sample_weights: Vector = None) -> Self:
        if sample_weights is None:
            sample_weights = np.ones(len(y))
        sample_weights = Vector(sample_weights)

        self.classes_ = np.unique(y)
        self.root = self._build_tree(X, y, 0, sample_weights)
        self._fitted = True
        return self

    def _stopping_criteria(self, y: Vector, depth: int) -> bool:
        if len(np.unique(y)) == 1:
            return True
        if depth == self.max_depth:
            return True
        if len(y) < self.min_samples_split:
            return True
        return False

    def _build_tree(
        self, X: Matrix, y: Vector, depth: int, sample_weights: Vector
    ) -> DecisionTreeNode:
        if self._stopping_criteria(y, depth):
            leaf_value = self._calculate_leaf_value(y, sample_weights)
            return DecisionTreeNode(value=leaf_value)

        feature_index, threshold = self._best_split(X, y, sample_weights)
        if feature_index is None:
            leaf_value = self._calculate_leaf_value(y, sample_weights)
            return DecisionTreeNode(value=leaf_value)

        left_indices = X[:, feature_index] <= threshold
        right_indices = X[:, feature_index] > threshold

        X_left, X_right = X[left_indices], X[right_indices]
        y_left, y_right = y[left_indices], y[right_indices]

        weights_left = sample_weights[left_indices]
        weights_right = sample_weights[right_indices]

        left_subtree = self._build_tree(X_left, y_left, depth + 1, weights_left)
        right_subtree = self._build_tree(X_right, y_right, depth + 1, weights_right)

        return DecisionTreeNode(
            feature_index=feature_index,
            threshold=threshold,
            left=left_subtree,
            right=right_subtree,
        )

    def _calculate_leaf_value(self, y: Vector, sample_weights: Vector) -> dict:
        classes, counts = np.unique(y, return_counts=True)
        weighted_counts = np.zeros_like(counts, dtype=np.float64)

        for i, cl in enumerate(classes):
            weighted_counts[i] = sample_weights[y == cl].sum()

        return dict(zip(classes, weighted_counts))

    def _best_split(
        self, X: Matrix, y: Vector, sample_weights: Vector
    ) -> Tuple[int, float]:
        best_gain = 0
        best_feature, best_threshold = None, None
        current_impurity = self._calculate_impurity(y, sample_weights)

        for feature_index in range(X.shape[1]):
            thresholds = np.unique(X[:, feature_index])
            for threshold in thresholds:
                left_indices = X[:, feature_index] <= threshold
                right_indices = X[:, feature_index] > threshold

                if (
                    np.sum(left_indices) < self.min_samples_split
                    or np.sum(right_indices) < self.min_samples_split
                ):
                    continue

                gain = self._calculate_gain(
                    y, sample_weights, left_indices, right_indices, current_impurity
                )
                if gain > best_gain:
                    best_gain = gain
                    best_feature = feature_index
                    best_threshold = threshold

        return best_feature, best_threshold

    def _calculate_impurity(self, y: Vector, sample_weights: Vector) -> float:
        if self.criterion == "gini":
            return self._calculate_gini(y, sample_weights)
        elif self.criterion == "entropy":
            return self._calculate_entropy(y, sample_weights)
        else:
            raise UnsupportedParameterError(self.criterion)

    def _calculate_gini(self, y: Vector, sample_weights: Vector) -> float:
        total_weight = np.sum(sample_weights)
        _, counts = np.unique(y, return_counts=True)
        weighted_counts = np.zeros_like(counts, dtype=np.float64)

        for i, cl in enumerate(np.unique(y)):
            weighted_counts[i] = np.sum(sample_weights[y == cl])

        prob = weighted_counts / total_weight
        gini = 1 - np.sum(prob**2)

        return gini

    def _calculate_entropy(self, y: Vector, sample_weights: Vector) -> float:
        total_weight = np.sum(sample_weights)
        _, counts = np.unique(y, return_counts=True)
        weighted_counts = np.zeros_like(counts, dtype=np.float64)

        for i, cl in enumerate(np.unique(y)):
            weighted_counts[i] = np.sum(sample_weights[y == cl])

        prob = weighted_counts / total_weight
        entropy = -np.sum(prob * np.log2(prob + np.finfo(float).eps))

        return entropy

    def _calculate_gain(
        self,
        y: Vector,
        sample_weights: Vector,
        left_indices: Vector,
        right_indices: Vector,
        current_impurity: float,
    ) -> float:
        left_weight = np.sum(sample_weights[left_indices])
        right_weight = np.sum(sample_weights[right_indices])
        total_weight = left_weight + right_weight

        weights_left = sample_weights[left_indices]
        weights_right = sample_weights[right_indices]

        left_impurity = self._calculate_impurity(y[left_indices], weights_left)
        right_impurity = self._calculate_impurity(y[right_indices], weights_right)

        weighted_impurity = (left_weight / total_weight) * left_impurity
        weighted_impurity += (right_weight / total_weight) * right_impurity
        gain = current_impurity - weighted_impurity

        return gain

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)
        preds = [self._predict_single(self.root, x) for x in X]

        return Vector(preds)

    def _predict_single(self, node: DecisionTreeNode, x: Vector) -> int:
        while node.value is None:
            if x[node.feature_index] <= node.threshold:
                node = node.left
            else:
                node = node.right

        return max(node.value, key=node.value.get)

    def predict_proba(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        proba_preds = [self._predict_proba_single(self.root, x) for x in X]

        return Matrix(proba_preds)

    def _predict_proba_single(self, node: DecisionTreeNode, x: Vector) -> Vector:
        while node.value is None:
            if x[node.feature_index] <= node.threshold:
                node = node.left
            else:
                node = node.right

        total_samples = sum(node.value.values())
        cl_probs = {cl: count / total_samples for cl, count in node.value.items()}

        probabilities = Matrix([cl_probs.get(cl, 0.0) for cl in self.classes_])
        return probabilities

    def score(self, X: Matrix, y: Vector, metric: Evaluator = Accuracy) -> float:
        preds = self.predict(X)
        return metric.score(y_true=y, y_pred=preds)



from typing import Literal, Tuple
import numpy as np

from luma.interface.typing import Tensor
from luma.interface.exception import UnsupportedParameterError
from luma.neural.base import Layer


__all__ = (
    "_Pool1D",
    "_Pool2D",
    "_Pool3D",
    "_GlobalAvgPool1D",
    "_GlobalAvgPool2D",
    "_GlobalAvgPool3D",
    "_LpPool1D",
    "_LpPool2D",
    "_LpPool3D",
)


class _Pool1D(Layer):
    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        mode: Literal["max", "avg"] = "max",
        padding: Tuple[int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__()
        self.filter_size = filter_size
        self.stride = stride
        self.mode = mode
        self.padding = padding

        self.set_param_ranges(
            {
                "filter_size": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(3)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, width = X.shape

        pad_w, padded_w = self._get_padding_dim(width)

        out_width = 1 + (padded_w - self.filter_size) // self.stride
        out: Tensor = np.zeros((batch_size, channels, out_width))

        X_padded = np.pad(X, ((0, 0), (0, 0), (pad_w, pad_w)), mode="constant")

        for i in range(out_width):
            w_start, w_end = self._get_pooling_bounds(i)
            window = X_padded[:, :, w_start:w_end]

            if self.mode == "max":
                out[:, :, i] = np.max(window, axis=2)
            elif self.mode == "avg":
                out[:, :, i] = np.mean(window, axis=2)
            else:
                raise UnsupportedParameterError(self.mode)

        return out

    @Tensor.force_dim(3)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, out_width = d_out.shape
        pad_w, _ = self._get_padding_dim(X.shape[2])
        X_padded = np.pad(X, ((0, 0), (0, 0), (pad_w, pad_w)), mode="constant")
        dX_padded = np.zeros_like(X_padded)

        for i in range(out_width):
            w_start, w_end = self._get_pooling_bounds(i)
            window = X_padded[:, :, w_start:w_end]

            if self.mode == "max":
                max_vals = np.max(window, axis=2, keepdims=True)
                mask_ = window == max_vals
                dX_padded[:, :, w_start:w_end] += mask_ * d_out[:, :, i : i + 1]
            elif self.mode == "avg":
                dX_padded[:, :, w_start:w_end] += (
                    d_out[:, :, i : i + 1] / self.filter_size
                )

        if pad_w > 0:
            self.dX = dX_padded[:, :, pad_w:-pad_w]
        else:
            self.dX = dX_padded

        return self.dX

    def _get_padding_dim(self, width: int) -> Tuple[int, ...]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 1:
                raise ValueError("Padding tuple must have exactly one value.")
            pad_w = self.padding[0]

        elif isinstance(self.padding, int):
            pad_w = self.padding
        elif self.padding == "same":
            pad_w = (self.filter_size - 1) // 2
        elif self.padding == "valid":
            pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_w = width + 2 * pad_w
        return pad_w, padded_w

    def _get_pooling_bounds(self, cur_w: int) -> Tuple[int, int]:
        w_start = cur_w * self.stride
        w_end = w_start + self.filter_size

        return w_start, w_end

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, width = in_shape
        _, padded_w = self._get_padding_dim(width)

        out_width = 1 + (padded_w - self.filter_size) // self.stride
        return (batch_size, channels, out_width)


class _Pool2D(Layer):
    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        mode: Literal["max", "avg"] = "max",
        padding: Tuple[int, int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__()
        self.filter_size = filter_size
        self.stride = stride
        self.mode = mode
        self.padding = padding

        self.set_param_ranges(
            {
                "filter_size": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(4)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, height, width = X.shape

        pad_h, pad_w, padded_h, padded_w = self._get_padding_dim(height, width)

        out_height = 1 + (padded_h - self.filter_size) // self.stride
        out_width = 1 + (padded_w - self.filter_size) // self.stride
        out: Tensor = np.zeros((batch_size, channels, out_height, out_width))

        X_padded = np.pad(
            X, ((0, 0), (0, 0), (pad_h, pad_h), (pad_w, pad_w)), mode="constant"
        )

        for i in range(out_height):
            for j in range(out_width):
                h_start, h_end, w_start, w_end = self._get_pooling_bounds(i, j)
                window = X_padded[:, :, h_start:h_end, w_start:w_end]

                if self.mode == "max":
                    out[:, :, i, j] = np.max(window, axis=(2, 3))
                elif self.mode == "avg":
                    out[:, :, i, j] = np.mean(window, axis=(2, 3))
                else:
                    raise UnsupportedParameterError(self.mode)

        return out

    @Tensor.force_dim(4)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, out_height, out_width = d_out.shape
        pad_h, pad_w, _, _ = self._get_padding_dim(X.shape[2], X.shape[3])
        X_padded = np.pad(
            X, ((0, 0), (0, 0), (pad_h, pad_h), (pad_w, pad_w)), mode="constant"
        )
        dX_padded = np.zeros_like(X_padded)

        for i in range(out_height):
            for j in range(out_width):
                h_start, h_end, w_start, w_end = self._get_pooling_bounds(i, j)
                window = X_padded[:, :, h_start:h_end, w_start:w_end]

                if self.mode == "max":
                    max_vals = np.max(window, axis=(2, 3), keepdims=True)
                    mask_ = window == max_vals
                    dX_padded[:, :, h_start:h_end, w_start:w_end] += (
                        mask_ * d_out[:, :, i : i + 1, j : j + 1]
                    )
                elif self.mode == "avg":
                    dX_padded[:, :, h_start:h_end, w_start:w_end] += d_out[
                        :, :, i : i + 1, j : j + 1
                    ] / (self.filter_size**2)

        if pad_h > 0 or pad_w > 0:
            self.dX = dX_padded[:, :, pad_h:-pad_h, pad_w:-pad_w]
        else:
            self.dX = dX_padded

        return self.dX

    def _get_padding_dim(self, height: int, width: int) -> Tuple[int, ...]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 2:
                raise ValueError("Padding tuple must have exactly two values.")
            pad_h, pad_w = self.padding
        elif isinstance(self.padding, int):
            pad_h = pad_w = self.padding

        elif self.padding == "same":
            pad_h = pad_w = (self.filter_size - 1) // 2
        elif self.padding == "valid":
            pad_h = pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_h = height + 2 * pad_h
        padded_w = width + 2 * pad_w

        return pad_h, pad_w, padded_h, padded_w

    def _get_pooling_bounds(self, cur_h: int, cur_w: int) -> Tuple[int, ...]:
        h_start = cur_h * self.stride
        w_start = cur_w * self.stride

        h_end = h_start + self.filter_size
        w_end = w_start + self.filter_size

        return h_start, h_end, w_start, w_end

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, height, width = in_shape
        _, _, padded_h, padded_w = self._get_padding_dim(height, width)

        out_height = 1 + (padded_h - self.filter_size) // self.stride
        out_width = 1 + (padded_w - self.filter_size) // self.stride

        return (batch_size, channels, out_height, out_width)


class _Pool3D(Layer):
    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        mode: Literal["max", "avg"] = "max",
        padding: Tuple[int, int, int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__()
        self.filter_size = filter_size
        self.stride = stride
        self.mode = mode
        self.padding = padding

        self.set_param_ranges(
            {
                "filter_size": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(5)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, depth, height, width = X.shape

        pad_d, pad_h, pad_w, padded_d, padded_h, padded_w = self._get_padding_dim(
            depth, height, width
        )

        out_depth = 1 + (padded_d - self.filter_size) // self.stride
        out_height = 1 + (padded_h - self.filter_size) // self.stride
        out_width = 1 + (padded_w - self.filter_size) // self.stride
        out: Tensor = np.zeros(
            (
                batch_size,
                channels,
                out_depth,
                out_height,
                out_width,
            )
        )
        X_padded = np.pad(
            X,
            ((0, 0), (0, 0), (pad_d, pad_d), (pad_h, pad_h), (pad_w, pad_w)),
            mode="constant",
        )

        for i in range(out_depth):
            for j in range(out_height):
                for k in range(out_width):
                    d_start, d_end, h_start, h_end, w_start, w_end = (
                        self._get_pooling_bounds(i, j, k)
                    )
                    window = X_padded[
                        :,
                        :,
                        d_start:d_end,
                        h_start:h_end,
                        w_start:w_end,
                    ]
                    if self.mode == "max":
                        out[:, :, i, j, k] = np.max(window, axis=(2, 3, 4))
                    elif self.mode == "avg":
                        out[:, :, i, j, k] = np.mean(window, axis=(2, 3, 4))
                    else:
                        raise UnsupportedParameterError(self.mode)

        return out

    @Tensor.force_dim(5)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, out_depth, out_height, out_width = d_out.shape
        pad_d, pad_h, pad_w, _, _, _ = self._get_padding_dim(
            X.shape[2], X.shape[3], X.shape[4]
        )
        X_padded = np.pad(
            X,
            ((0, 0), (0, 0), (pad_d, pad_d), (pad_h, pad_h), (pad_w, pad_w)),
            mode="constant",
        )
        dX_padded = np.zeros_like(X_padded)

        for i in range(out_depth):
            for j in range(out_height):
                for k in range(out_width):
                    d_start, d_end, h_start, h_end, w_start, w_end = (
                        self._get_pooling_bounds(i, j, k)
                    )
                    window = X_padded[
                        :,
                        :,
                        d_start:d_end,
                        h_start:h_end,
                        w_start:w_end,
                    ]
                    if self.mode == "max":
                        max_vals = np.max(window, axis=(2, 3, 4), keepdims=True)
                        mask_ = window == max_vals
                        dX_padded[
                            :, :, d_start:d_end, h_start:h_end, w_start:w_end
                        ] += (mask_ * d_out[:, :, i : i + 1, j : j + 1, k : k + 1])
                    elif self.mode == "avg":
                        dX_padded[
                            :, :, d_start:d_end, h_start:h_end, w_start:w_end
                        ] += d_out[:, :, i : i + 1, j : j + 1, k : k + 1] / (
                            self.filter_size**3
                        )

        if pad_d > 0 or pad_h > 0 or pad_w > 0:
            self.dX = dX_padded[:, :, pad_d:-pad_d, pad_h:-pad_h, pad_w:-pad_w]
        else:
            self.dX = dX_padded

        return self.dX

    def _get_padding_dim(
        self,
        depth: int,
        height: int,
        width: int,
    ) -> Tuple[int, ...]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 3:
                raise ValueError("Padding tuple must have exactly three values.")
            pad_d, pad_h, pad_w = self.padding

        elif isinstance(self.padding, int):
            pad_d = pad_h = pad_w = self.padding
        elif self.padding == "same":
            pad_d = pad_h = pad_w = (self.filter_size - 1) // 2
        elif self.padding == "valid":
            pad_d = pad_h = pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_d = depth + 2 * pad_d
        padded_h = height + 2 * pad_h
        padded_w = width + 2 * pad_w

        return pad_d, pad_h, pad_w, padded_d, padded_h, padded_w

    def _get_pooling_bounds(
        self, cur_d: int, cur_h: int, cur_w: int
    ) -> Tuple[int, ...]:
        d_start = cur_d * self.stride
        h_start = cur_h * self.stride
        w_start = cur_w * self.stride

        d_end = d_start + self.filter_size
        h_end = h_start + self.filter_size
        w_end = w_start + self.filter_size

        return d_start, d_end, h_start, h_end, w_start, w_end

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, depth, height, width = in_shape
        _, _, _, padded_d, padded_h, padded_w = self._get_padding_dim(
            depth, height, width
        )
        out_depth = 1 + (padded_d - self.filter_size) // self.stride
        out_height = 1 + (padded_h - self.filter_size) // self.stride
        out_width = 1 + (padded_w - self.filter_size) // self.stride

        return (batch_size, channels, out_depth, out_height, out_width)


class _GlobalAvgPool1D(Layer):
    def __init__(self) -> None:
        super().__init__()

    @Tensor.force_dim(3)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        out = np.mean(X, axis=2, keepdims=True)

        return out

    @Tensor.force_dim(3)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, width = X.shape

        dX = np.zeros_like(X)
        dX += d_out / width
        self.dX = dX
        return self.dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, _ = in_shape
        return (batch_size, channels, 1)


class _GlobalAvgPool2D(Layer):
    def __init__(self) -> None:
        super().__init__()

    @Tensor.force_dim(4)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        out = np.mean(X, axis=(2, 3), keepdims=True)

        return out

    @Tensor.force_dim(4)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, height, width = X.shape

        dX = np.zeros_like(X)
        dX += d_out / (height * width)
        self.dX = dX
        return self.dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, _, _ = in_shape
        return (batch_size, channels, 1, 1)


class _GlobalAvgPool3D(Layer):
    def __init__(self) -> None:
        super().__init__()

    @Tensor.force_dim(5)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        out = np.mean(X, axis=(2, 3, 4), keepdims=True)

        return out

    @Tensor.force_dim(5)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, depth, height, width = X.shape

        dX = np.zeros_like(X)
        dX += d_out / (depth * height * width)
        self.dX = dX
        return self.dX

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, _, _, _ = in_shape
        return (batch_size, channels, 1, 1, 1)


class _LpPool1D(Layer):
    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        p: float = 2.0,
        padding: Tuple[int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__()
        self.filter_size = filter_size
        self.stride = stride
        self.p = p
        self.padding = padding

        self.set_param_ranges(
            {
                "filter_size": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
                "p": ("0<,+inf", float),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(3)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, width = X.shape

        pad_w, padded_w = self._get_padding_dim(width)

        out_width = 1 + (padded_w - self.filter_size) // self.stride
        out: Tensor = np.zeros((batch_size, channels, out_width))

        X_padded = np.pad(X, ((0, 0), (0, 0), (pad_w, pad_w)), mode="constant")

        for i in range(out_width):
            w_start, w_end = self._get_pooling_bounds(i)
            window = X_padded[:, :, w_start:w_end]

            out[:, :, i] = np.power(
                np.sum(np.power(np.abs(window), self.p), axis=2), 1 / self.p
            )

        return out

    @Tensor.force_dim(3)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, width = X.shape

        pad_w, _ = self._get_padding_dim(width)
        X_padded = np.pad(X, ((0, 0), (0, 0), (pad_w, pad_w)), mode="constant")
        dX_padded = np.zeros_like(X_padded)

        out_width = d_out.shape[2]

        for i in range(out_width):
            w_start, w_end = self._get_pooling_bounds(i)
            window = X_padded[:, :, w_start:w_end]

            window_p_norm = np.power(
                np.sum(np.power(np.abs(window), self.p), axis=2), 1 / self.p
            )

            gradient = (np.power(np.abs(window), self.p - 1) * np.sign(window)) / (
                np.power(window_p_norm, self.p - 1, keepdims=True) + 1e-12
            )

            dX_padded[:, :, w_start:w_end] += gradient * d_out[:, :, i : i + 1]

        if pad_w > 0:
            self.dX = dX_padded[:, :, pad_w:-pad_w]
        else:
            self.dX = dX_padded

        return self.dX

    def _get_padding_dim(self, width: int) -> Tuple[int, int]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 1:
                raise ValueError("Padding tuple must have exactly one value.")
            pad_w = self.padding[0]

        elif isinstance(self.padding, int):
            pad_w = self.padding
        elif self.padding == "same":
            pad_w = (self.filter_size - 1) // 2
        elif self.padding == "valid":
            pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_w = width + 2 * pad_w
        return pad_w, padded_w

    def _get_pooling_bounds(self, cur_w: int) -> Tuple[int, int]:
        w_start = cur_w * self.stride
        w_end = w_start + self.filter_size

        return w_start, w_end

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, width = in_shape
        _, padded_w = self._get_padding_dim(width)

        out_width = 1 + (padded_w - self.filter_size) // self.stride
        return (batch_size, channels, out_width)


class _LpPool2D(Layer):
    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        p: float = 2.0,
        padding: Tuple[int, int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__()
        self.filter_size = filter_size
        self.stride = stride
        self.p = p
        self.padding = padding

        self.set_param_ranges(
            {
                "filter_size": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
                "p": ("0<,+inf", float),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(4)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, height, width = X.shape

        pad_h, pad_w, padded_h, padded_w = self._get_padding_dim(height, width)

        out_height = 1 + (padded_h - self.filter_size) // self.stride
        out_width = 1 + (padded_w - self.filter_size) // self.stride
        out: Tensor = np.zeros((batch_size, channels, out_height, out_width))

        X_padded = np.pad(
            X, ((0, 0), (0, 0), (pad_h, pad_h), (pad_w, pad_w)), mode="constant"
        )

        for i in range(out_height):
            for j in range(out_width):
                h_start, h_end, w_start, w_end = self._get_pooling_bounds(i, j)
                window = X_padded[:, :, h_start:h_end, w_start:w_end]

                out[:, :, i, j] = np.power(
                    np.sum(np.power(np.abs(window), self.p), axis=(2, 3)), 1 / self.p
                )

        return out

    @Tensor.force_dim(4)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, height, width = X.shape

        pad_h, pad_w, _, _ = self._get_padding_dim(height, width)
        X_padded = np.pad(
            X, ((0, 0), (0, 0), (pad_h, pad_h), (pad_w, pad_w)), mode="constant"
        )
        dX_padded = np.zeros_like(X_padded)

        out_height, out_width = d_out.shape[2], d_out.shape[3]

        for i in range(out_height):
            for j in range(out_width):
                h_start, h_end, w_start, w_end = self._get_pooling_bounds(i, j)
                window = X_padded[:, :, h_start:h_end, w_start:w_end]

                window_p_norm = np.power(
                    np.sum(np.power(np.abs(window), self.p), axis=(2, 3)), 1 / self.p
                )

                gradient = (np.power(np.abs(window), self.p - 1) * np.sign(window)) / (
                    np.power(window_p_norm, self.p - 1) + 1e-12
                )

                dX_padded[:, :, h_start:h_end, w_start:w_end] += (
                    gradient * d_out[:, :, i : i + 1, j : j + 1]
                )

        if pad_h > 0 or pad_w > 0:
            self.dX = dX_padded[:, :, pad_h:-pad_h, pad_w:-pad_w]
        else:
            self.dX = dX_padded

        return self.dX

    def _get_padding_dim(self, height: int, width: int) -> Tuple[int, ...]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 2:
                raise ValueError("Padding tuple must have exactly two values.")
            pad_h, pad_w = self.padding
        elif isinstance(self.padding, int):
            pad_h = pad_w = self.padding

        elif self.padding == "same":
            pad_h = pad_w = (self.filter_size - 1) // 2
        elif self.padding == "valid":
            pad_h = pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_h = height + 2 * pad_h
        padded_w = width + 2 * pad_w

        return pad_h, pad_w, padded_h, padded_w

    def _get_pooling_bounds(self, cur_h: int, cur_w: int) -> Tuple[int, ...]:
        h_start = cur_h * self.stride
        w_start = cur_w * self.stride

        h_end = h_start + self.filter_size
        w_end = w_start + self.filter_size

        return h_start, h_end, w_start, w_end

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, height, width = in_shape
        _, _, padded_h, padded_w = self._get_padding_dim(height, width)

        out_height = 1 + (padded_h - self.filter_size) // self.stride
        out_width = 1 + (padded_w - self.filter_size) // self.stride

        return (batch_size, channels, out_height, out_width)


class _LpPool3D(Layer):
    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        p: float = 2.0,
        padding: Tuple[int, int, int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__()
        self.filter_size = filter_size
        self.stride = stride
        self.p = p
        self.padding = padding

        self.set_param_ranges(
            {
                "filter_size": ("0<,+inf", int),
                "stride": ("0<,+inf", int),
                "p": ("0<,+inf", float),
            }
        )
        self.check_param_ranges()

    @Tensor.force_dim(5)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        _ = is_train
        self.input_ = X
        batch_size, channels, depth, height, width = X.shape

        pad_d, pad_h, pad_w, padded_d, padded_h, padded_w = self._get_padding_dim(
            depth, height, width
        )

        out_depth = 1 + (padded_d - self.filter_size) // self.stride
        out_height = 1 + (padded_h - self.filter_size) // self.stride
        out_width = 1 + (padded_w - self.filter_size) // self.stride
        out: Tensor = np.zeros((batch_size, channels, out_depth, out_height, out_width))

        X_padded = np.pad(
            X,
            ((0, 0), (0, 0), (pad_d, pad_d), (pad_h, pad_h), (pad_w, pad_w)),
            mode="constant",
        )

        for i in range(out_depth):
            for j in range(out_height):
                for k in range(out_width):
                    d_start, d_end, h_start, h_end, w_start, w_end = (
                        self._get_pooling_bounds(i, j, k)
                    )
                    window = X_padded[:, :, d_start:d_end, h_start:h_end, w_start:w_end]

                    out[:, :, i, j, k] = np.power(
                        np.sum(np.power(np.abs(window), self.p), axis=(2, 3, 4)),
                        1 / self.p,
                    )

        return out

    @Tensor.force_dim(5)
    def backward(self, d_out: Tensor) -> Tensor:
        X = self.input_
        _, _, depth, height, width = X.shape

        pad_d, pad_h, pad_w, _, _, _ = self._get_padding_dim(depth, height, width)
        X_padded = np.pad(
            X,
            ((0, 0), (0, 0), (pad_d, pad_d), (pad_h, pad_h), (pad_w, pad_w)),
            mode="constant",
        )
        dX_padded = np.zeros_like(X_padded)

        out_depth, out_height, out_width = (
            d_out.shape[2],
            d_out.shape[3],
            d_out.shape[4],
        )
        for i in range(out_depth):
            for j in range(out_height):
                for k in range(out_width):
                    d_start, d_end, h_start, h_end, w_start, w_end = (
                        self._get_pooling_bounds(i, j, k)
                    )
                    window = X_padded[:, :, d_start:d_end, h_start:h_end, w_start:w_end]

                    window_p_norm = np.power(
                        np.sum(np.power(np.abs(window), self.p), axis=(2, 3, 4)),
                        1 / self.p,
                    )

                    gradient = (
                        np.power(np.abs(window), self.p - 1) * np.sign(window)
                    ) / (np.power(window_p_norm, self.p - 1, keepdims=True) + 1e-12)

                    dX_padded[:, :, d_start:d_end, h_start:h_end, w_start:w_end] += (
                        gradient * d_out[:, :, i : i + 1, j : j + 1, k : k + 1]
                    )

        if pad_d > 0 or pad_h > 0 or pad_w > 0:
            self.dX = dX_padded[:, :, pad_d:-pad_d, pad_h:-pad_h, pad_w:-pad_w]
        else:
            self.dX = dX_padded

        return self.dX

    def _get_padding_dim(
        self,
        depth: int,
        height: int,
        width: int,
    ) -> Tuple[int, ...]:
        if isinstance(self.padding, tuple):
            if len(self.padding) != 3:
                raise ValueError("Padding tuple must have exactly three values.")
            pad_d, pad_h, pad_w = self.padding
        elif isinstance(self.padding, int):
            pad_d = pad_h = pad_w = self.padding

        elif self.padding == "same":
            pad_d = pad_h = pad_w = (self.filter_size - 1) // 2
        elif self.padding == "valid":
            pad_d = pad_h = pad_w = 0
        else:
            raise UnsupportedParameterError(self.padding)

        padded_d = depth + 2 * pad_d
        padded_h = height + 2 * pad_h
        padded_w = width + 2 * pad_w

        return pad_d, pad_h, pad_w, padded_d, padded_h, padded_w

    def _get_pooling_bounds(
        self, cur_d: int, cur_h: int, cur_w: int
    ) -> Tuple[int, ...]:
        d_start = cur_d * self.stride
        h_start = cur_h * self.stride
        w_start = cur_w * self.stride

        d_end = d_start + self.filter_size
        h_end = h_start + self.filter_size
        w_end = w_start + self.filter_size

        return d_start, d_end, h_start, h_end, w_start, w_end

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        batch_size, channels, depth, height, width = in_shape
        _, _, _, padded_d, padded_h, padded_w = self._get_padding_dim(
            depth, height, width
        )

        out_depth = 1 + (padded_d - self.filter_size) // self.stride
        out_height = 1 + (padded_h - self.filter_size) // self.stride
        out_width = 1 + (padded_w - self.filter_size) // self.stride

        return (batch_size, channels, out_depth, out_height, out_width)



import numpy as np

from luma.interface.typing import Vector, Matrix, Tensor, TensorLike


__all__ = ("KaimingInit", "XavierInit")


class Initializer:
    @classmethod
    def __class_alias__(cls) -> None: ...


class KaimingInit(Initializer):
    def __init__(self, random_state: int = None) -> None:
        super().__init__()
        self.rs_ = np.random.RandomState(random_state)

    def init_nd(
        self,
        input_size: int,
        output_size: int,
        *shape: int,
    ) -> TensorLike:
        stddev = np.sqrt(2.0 / (input_size * np.prod(Vector(shape))))
        return self.rs_.normal(
            0.0,
            stddev,
            (input_size, output_size, *shape),
        )


class XavierInit(Initializer):
    def __init__(self, random_state: int = None) -> None:
        super().__init__()
        self.rs_ = np.random.RandomState(random_state)

    def init_nd(
        self,
        input_size: int,
        output_size: int,
        *shape: int,
    ) -> TensorLike:
        stddev = np.sqrt(
            2.0
            / (
                input_size * np.prod(Vector(shape))
                + output_size * np.prod(Vector(shape))
            )
        )
        return self.rs_.normal(
            0.0,
            stddev,
            (input_size, output_size, *shape),
        )



from typing import Self, Tuple
from scipy.spatial.distance import cdist
import numpy as np

from luma.core.super import Transformer
from luma.interface.typing import Matrix, Vector
from luma.interface.exception import NotFittedError


__all__ = "LocalOutlierFactor"


class LocalOutlierFactor(Transformer, Transformer.Both):
    """
    The Local Outlier Factor (LOF) algorithm identifies outliers by comparing
    the local density of a data point with the densities of its neighbors.
    Points with significantly lower density than their neighbors are considered
    outliers. LOF scores greater than 1 indicate potential outliers. It is
    effective in datasets with varying densities and does not require a prior
    assumption about the data distribution.

    Parameters
    ----------
    `n_neighbors` : int, default=10
        Number of neighbors to estimate the local densities
    `threshold` : float, default=1.5
        Cutoff value deciding whether a point is considered an outlier

    Examples
    --------
    >>> lof = LocalOutlierFactor()
    >>> lof.fit(X)

    Getting LOF Scores
    >>> lof_scores = lof.get_scores(X)

    Filtering Dataset
    >>> X_, y_ = lof.filter(X, y) # or lof.transform(X, y)

    """

    def __init__(self, n_neighbors: int = 10, threshold: float = 1.5):
        self.n_neighbors = n_neighbors
        self.threshold = threshold
        self.lrd_ = None
        self.neighbors_ = None
        self._fitted = False

        self.set_param_ranges(
            {"n_neighbors": ("0<,+inf", int), "threshold": ("0,+inf", None)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, _=None) -> Self:
        n_samples = len(X)
        self.lrd_ = np.zeros(n_samples)
        self.neighbors_ = []

        for i in range(n_samples):
            distances = cdist([X[i]], X)[0]
            neighbors_idx = np.argsort(distances)[1 : self.n_neighbors + 1]
            neighbors = X[neighbors_idx]

            self.lrd_[i] = self._local_reachability_density(X[i], neighbors)
            self.neighbors_.append(neighbors_idx)

        self._fitted = True
        return self

    def _k_distance(self, x: Matrix, neighbors: Matrix) -> Matrix:
        sorted_ = np.sort(cdist([x], neighbors)[0])
        return sorted_[self.n_neighbors - 1]

    def _reachability_distance(
        self, x: Matrix, neighbor: Matrix, neighbors: Matrix
    ) -> Matrix:
        k_dist_neighbor = self._k_distance(neighbor, neighbors)
        return max(k_dist_neighbor, np.linalg.norm(x - neighbor))

    def _local_reachability_density(self, x: Matrix, neighbors: Matrix) -> Matrix:
        reach_distances = [
            self._reachability_distance(x, neighbor, neighbors)
            for neighbor in neighbors
        ]
        return 1 / (
            np.sum(reach_distances) / len(neighbors) if len(neighbors) > 0 else 1
        )

    def get_scores(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        lof = np.zeros(len(X))
        for i in range(len(X)):
            lrd_sum = sum(self.lrd_[idx] for idx in self.neighbors_[i])
            lof[i] = lrd_sum / (self.n_neighbors * self.lrd_[i])

        return lof

    def filter(self, X: Matrix, y: Matrix) -> Tuple[Matrix, Vector]:
        lof_scores = self.get_scores(X)
        condition = lof_scores < self.threshold

        return X[condition], y[condition]

    def transform(self, X: Matrix, y: Vector) -> Tuple[Matrix, Vector]:
        return self.filter(X, y)

    def fit_transform(self, X: Matrix, y: Vector) -> Tuple[Matrix, Vector]:
        self.fit(X)
        return self.transform(X, y)



from typing import Self
from scipy.spatial import distance_matrix
import numpy as np

from luma.core.super import Estimator, Supervised, Evaluator
from luma.interface.typing import Matrix
from luma.interface.util import NearestNeighbors
from luma.interface.exception import NotFittedError
from luma.metric.classification import Accuracy


__all__ = ("KNNClassifier", "AdaptiveKNNClassifier", "WeightedKNNClassifier")


class KNNClassifier(Estimator, Supervised):
    """
    The K-Nearest Neighbors (KNN) classifier is a simple and intuitive
    machine learning algorithm that classifies a new data point based on
    the majority class among its closest neighbors. It operates by
    calculating the distance (such as Euclidean distance) between the
    new point and all points in the training set, identifying the 'k'
    nearest neighbors, and then performing a majority vote among these
    neighbors to determine the class label.

    Parameters
    ----------
    `n_neighbors` : int, default=5
        Number of neighbors to be considered close

    """

    def __init__(self, n_neighbors: int = 5):
        self.n_neighbors = n_neighbors
        self._neighbors = None
        self._y = None
        self._fitted = False

        self.set_param_ranges({"n_neighbors": ("0,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Matrix) -> Self:
        self._neighbors = NearestNeighbors(X, self.n_neighbors)
        self._y = y
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        predictions = []

        for x in X:
            distances = np.linalg.norm(self._neighbors.data - x, axis=1)
            nearest_neighbor_ids = np.argsort(distances)[: self.n_neighbors]
            nearest_neighbor_labels = self._y[nearest_neighbor_ids]

            most_common_label = np.bincount(nearest_neighbor_labels).argmax()
            predictions.append(most_common_label)

        return Matrix(predictions)

    def predict_proba(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        proba = []

        for x in X:
            distances = np.linalg.norm(self._neighbors.data - x, axis=1)
            nearest_neighbor_ids = np.argsort(distances)[: self.n_neighbors]
            nearest_neighbor_labels = self._y[nearest_neighbor_ids]

            class_votes = np.bincount(
                nearest_neighbor_labels, minlength=np.max(self._y) + 1
            )

            class_probabilities = class_votes / self.n_neighbors
            proba.append(class_probabilities)

        return Matrix(proba)

    def score(self, X: Matrix, y: Matrix, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)


class AdaptiveKNNClassifier(Estimator, Supervised):
    """
    The Adaptive K-Nearest Neighbors (AdaKNN) Classifier is an extension
    of the conventional KNN algorithm for classification. It classifies a
    new data point based on the majority class among its neighbors, where
    the number of neighbors (k) is adaptively determined based on the local
    density of the training data. This adaptive approach allows the algorithm
    to be more flexible and effective, especially in datasets with varying
    densities.

    Parameters
    ----------
    `n_density` : int, default=10
        Number of nearest neighbors to estimate the local density
    `min_neighbors` : int, default=5
        Minimum number of neighbors to be considered for averaging
    `max_neighbors` : int, default=20
        Maximum number of neighbors to be considered

    """

    def __init__(
        self, n_density: int = 10, min_neighbors: int = 5, max_neighbors: int = 20
    ):
        self.n_density = n_density
        self.min_neighbors = min_neighbors
        self.max_neighbors = max_neighbors
        self._X = None
        self._y = None
        self._fitted = False

        self.set_param_ranges(
            {
                "n_density": ("0<,+int", int),
                "min_neighbors": ("0<,+int", int),
                "max_neighbors": ("0<,+int", int),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Matrix) -> Self:
        self._X = X
        self._y = y

        self.dist_matrix = distance_matrix(X, X)
        self.local_density = np.sort(self.dist_matrix, axis=1)
        self.local_density = self.local_density[:, self.n_density]

        self.k_values = np.clip(
            np.ceil(self.max_neighbors / self.local_density),
            self.min_neighbors,
            self.max_neighbors,
        ).astype(int)

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        dist_matrix = distance_matrix(X, self._X)

        predictions = []
        for point_distances in dist_matrix:
            nearest_indices = np.argsort(point_distances)
            votes = np.zeros(len(np.unique(self._y)))

            for idx, _ in enumerate(point_distances[nearest_indices]):
                if idx >= self.k_values[nearest_indices[idx]]:
                    break
                votes[self._y[nearest_indices[idx]]] += 1

            most_common = np.argmax(votes)
            predictions.append(most_common)

        return Matrix(predictions)

    def predict_proba(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        dist_matrix = distance_matrix(X, self._X)

        proba_predictions = []
        for i, point_distances in enumerate(dist_matrix):
            nearest_indices = np.argsort(point_distances)
            k_value = min(self.k_values[i], len(nearest_indices))
            nearest_indices = nearest_indices[:k_value]

            votes = np.zeros(np.max(self._y) + 1)
            for idx in nearest_indices:
                votes[self._y[idx]] += 1

            class_probabilities = votes / votes.sum()
            proba_predictions.append(class_probabilities)

        return Matrix(proba_predictions)

    def score(self, X: Matrix, y: Matrix, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)


class WeightedKNNClassifier(Estimator, Supervised):
    """
    The Weighted KNN Classifier is a variation of the k-Nearest Neighbors algorithm
    where neighbors contribute to the classification decision based on their distance
    from the query point. Closer neighbors have more influence as they are assigned
    higher weights, typically using inverse distance weighting. This approach enhances
    prediction accuracy, especially in unevenly distributed datasets.

    Parameters
    ----------
    `n_neighbors` : int, default=5
        Number of neighbors to be considered close

    """

    def __init__(self, n_neighbors: int = 5):
        self.n_neighbors = n_neighbors
        self._X = None
        self._y = None
        self._fitted = False

        self.set_param_ranges({"n_neighbors": ("0,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Matrix) -> Self:
        self._X = X
        self._y = y

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        predictions = np.zeros(X.shape[0])

        for i, x in enumerate(X):
            distances = np.linalg.norm(self._X - x, axis=1)
            nearest_neighbors = np.argsort(distances)[: self.n_neighbors]

            weights = 1 / (distances[nearest_neighbors] + 1e-5)
            weighted_votes = np.zeros(np.unique(self._y).shape[0])
            for idx, neighbor in enumerate(nearest_neighbors):
                weighted_votes[self._y[neighbor]] += weights[idx]

            predictions[i] = np.argmax(weighted_votes)

        return predictions.astype(int)

    def predict_proba(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        proba_predictions = []

        for x in X:
            distances = np.linalg.norm(self._X - x, axis=1)
            nearest_neighbor_indices = np.argsort(distances)[: self.n_neighbors]

            weights = 1 / (distances[nearest_neighbor_indices] + 1e-5)
            all_classes = np.unique(self._y)
            weighted_votes = np.zeros(len(all_classes))

            for i, cl in enumerate(all_classes):
                class_weights = weights[self._y[nearest_neighbor_indices] == cl]
                weighted_votes[i] = np.sum(class_weights)

            class_probabilities = weighted_votes / np.sum(weighted_votes)
            proba_predictions.append(class_probabilities)

        return Matrix(proba_predictions)

    def score(self, X: Matrix, y: Matrix, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)



import platform

arc = platform.machine()

MLX_TOGGLE: bool = False

try:
    if MLX_TOGGLE:
        import mlx as _
except ModuleNotFoundError as err:
    if arc != "arm64":
        print(f"Your python environment is not running on 'arm64', but on '{arc}'")
        print(
            f"Run luma/arm.sh to setup an ARM supported Python environment",
            f"and install MLX via: 'pip install mlx'.",
        )
    else:
        print(err)



from typing import Optional
from scipy.spatial.distance import pdist, squareform
import matplotlib.pyplot as plt
import numpy as np

from luma.core.super import Evaluator, Visualizer
from luma.interface.util import SilhouetteUtil, DBUtil
from luma.interface.typing import Matrix, Vector, ClassType


__all__ = ("SilhouetteCoefficient", "DaviesBouldin", "Inertia")


class SilhouetteCoefficient(Evaluator, Visualizer):
    """
    The Silhouette Coefficient is a measure used to evaluate the quality of
    clusters in a dataset. It calculates how similar each data point is to its
    own cluster compared to other clusters. The coefficient ranges from -1
    (poorly clustered) to +1 (well clustered), with values near 0 indicating
    overlapping clusters. High average scores across a dataset suggest clear,
    well-separated clusters.

    Parameters
    ----------
    `data` : Matrix
        Original data
    `labels` : Vector
        Labels assigned by clustering estimator

    Examples
    --------
    With Instantiation
    ```py
        sil = SilhouetteCoefficient(data, labels)
        score = sil.score(data, labels) # compute() is a static method
        sil.plot(...)
    ```
    Without Instantiation
    ```py
        score = SilhouetteCoefficient.score(data, labels)
        SilhouetteCoefficient.plot(...) # Error; plot() is an instance method
    ```
    """

    def __init__(self, data: Matrix, labels: Vector) -> None:
        self.data = data
        self.labels = labels
        self.dist = squareform(pdist(self.data))

    @staticmethod
    def score(data: Matrix, labels: Matrix) -> Matrix[float]:
        scores = []
        dist = squareform(pdist(data))
        for idx, label in enumerate(labels):
            util = SilhouetteUtil(idx, label, labels, dist)
            a = util.avg_dist_within
            b = util.avg_dist_others
            score = (b - a) / max(a, b) if max(a, b) != 0 else 0
            scores.append(score)

        return np.mean(scores)

    def _individual_silhouette(self) -> Matrix:
        silhouette_values = []
        for idx, label in enumerate(self.labels):
            util = SilhouetteUtil(idx, label, self.labels, self.dist)
            a = util.avg_dist_within
            b = util.avg_dist_others
            score = (b - a) / max(a, b) if max(a, b) != 0 else 0
            silhouette_values.append(score)

        return Matrix(silhouette_values)

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        sample_silhouette = self._individual_silhouette()
        y_lower = 10

        for i in range(len(np.unique(self.labels))):
            values = sample_silhouette[self.labels == i]
            values.sort()

            size_cluster_i = values.shape[0]
            y_upper = y_lower + size_cluster_i

            color = plt.cm.nipy_spectral(float(i) / len(set(self.labels)))
            ax.fill_betweenx(
                np.arange(y_lower, y_upper),
                0,
                values,
                facecolor=color,
                edgecolor=color,
                alpha=0.7,
            )
            y_lower = y_upper + 10

        ax.set_title("Silhouette Coefficient Plot")
        ax.set_xlabel("Silhouette coefficient values")
        ax.set_ylabel("Cluster label")
        ax.axvline(x=self.score(self.data, self.labels), color="red", linestyle="--")

        ax.set_yticks([])
        ax.set_xticks(np.arange(0.0, 1.2, 0.1))
        plt.tight_layout()

        if show:
            plt.show()
        return ax


@ClassType.non_instantiable()
class DaviesBouldin(Evaluator):
    """
    The Davies-Bouldin Index (DBI) is a metric for evaluating clustering
    algorithms. It compares the average distance within clusters to the
    distance between clusters. Lower DBI values indicate better clustering,
    with compact and well-separated clusters.

    Parameters
    ----------
    `data` : Matrix
        Original data
    `labels` : Vector
        Labels assigned by clustering estimator

    """

    @classmethod
    def score(cls, data: Matrix, labels: Vector) -> float:
        util = DBUtil(data=data, labels=labels)
        centroids = util.cluster_centroids
        scatter = util.within_cluster_scatter
        separation = util.separation

        n_clusters = len(centroids)
        db_values = np.zeros(n_clusters)

        for i in range(n_clusters):
            max_ratio = 0
            for j in range(n_clusters):
                if i == j:
                    continue
                ratio = (scatter[i] + scatter[j]) / separation[i, j]
                if ratio > max_ratio:
                    max_ratio = ratio

            db_values[i] = max_ratio
        db_index = np.mean(db_values)

        return db_index


@ClassType.non_instantiable()
class Inertia(Evaluator):
    """
    Inertia in clustering quantifies the compactness of clusters by measuring
    the sum of squared distances between data points and their respective
    cluster centroids. It serves as a key metric to evaluate the quality of
    cluster assignments in algorithms like K-means. A lower inertia value
    indicates more cohesive clusters, where data points are closer to their
    centroids. However, minimizing inertia excessively can lead to overfitting,
    with a large number of clusters.

    Parameters
    ----------
    `data` : Matrix
        Original data
    `centroids`: Matrix
        Centroids(or medoids for certain algorithm)

    """

    @classmethod
    def score(cls, data: Matrix, centroids: Matrix) -> float:
        sq_dist = (data[:, np.newaxis, :] - centroids[np.newaxis, :, :]) ** 2
        dist = np.sqrt(sq_dist.sum(axis=2))

        closest = np.argmin(dist, axis=1)
        inertia = sum((dist[i, closest[i]] ** 2) for i in range(len(data)))

        return inertia
